/*
 *	spDialog_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdarg.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spLabel.h>

#include <sp/spTopLevelP.h>
#include <sp/spParamFieldP.h>
#include <sp/spFrameP.h>
#include <sp/spGraphicsP.h>
#include <sp/spDialogBoxP.h>
#include <sp/spDialogP.h>

#include <gdk/gdkkeysyms.h>

static gint modalLoop(GtkWidget *dialog)
{
    gint result = 0;
#if !GTK_CHECK_VERSION(2,4,0)
    GtkWidget *prev_grabbed;
    
    spDebug(30, "modalLoop", "before gtk_widget_show\n");
    gtk_widget_show(dialog);
    spDebug(30, "modalLoop", "gtk_widget_show done\n");
    prev_grabbed = spAddGrabGtk(dialog);
#endif
	
    spDebug(30, "modalLoop", "before spMainLoopGtk\n");
    result = spMainLoopGtk(dialog, SP_TRUE);
	
#if !GTK_CHECK_VERSION(2,4,0)
    spAddGrabGtk(prev_grabbed);
    gtk_widget_hide(dialog);
#endif

    return result;
}

void spDialogPartInitArch(spComponent component)
{
    SpDialogArch(component).file_type_field = NULL;
    
    return;
}

#if !GTK_CHECK_VERSION(2,4,0)
static void mainQuit(GtkWidget *widget, spBool ok_flag)
{
    spRemoveGrabGtk(NULL);
    spDebug(30, "mainQuit", "call spMainQuitGtk\n");
    spMainQuitGtk();
    
    spDebug(30, "mainQuit", "done\n");
    
    return;
}
#endif

#if GTK_CHECK_VERSION(2,4,0)
spDialogResponse spPopupMessageBox(spComponent component, const char *message)
{
    GtkWindow *parent = NULL;
    GtkWidget *dialog = NULL;
    GtkMessageType message_type = GTK_MESSAGE_OTHER;
    GtkButtonsType buttons_type = GTK_BUTTONS_NONE;
    gint result;
    spDialogResponse response = SP_DR_CANCEL;

    if (spIsFrame(SpGetParent(component)) == SP_TRUE) {
	parent = GTK_WINDOW(SpParentFrameArch(component).toplevel);
    }
    
    if (SpDialogPart(component).dialog_type == SP_INFORMATION_DIALOG) {
	message_type = GTK_MESSAGE_INFO;
    } else if (SpDialogPart(component).dialog_type == SP_QUESTION_DIALOG) {
	message_type = GTK_MESSAGE_QUESTION;
    } else if (SpDialogPart(component).dialog_type == SP_WARNING_DIALOG) {
	message_type = GTK_MESSAGE_WARNING;
    } else if (SpDialogPart(component).dialog_type == SP_ERROR_DIALOG) {
	message_type = GTK_MESSAGE_ERROR;
    }
    
    if (SpDialogPart(component).button_type == SP_MB_OK) {
	buttons_type = GTK_BUTTONS_OK;
	response = SP_DR_OK;
    } else if (SpDialogPart(component).button_type == SP_MB_OK_CANCEL
	       || SpDialogPart(component).button_type == SP_MB_OK_CANCEL_APPLY) {
	buttons_type = GTK_BUTTONS_OK_CANCEL;
    } else if (SpDialogPart(component).button_type == SP_MB_YES_NO
	       || SpDialogPart(component).button_type == SP_MB_YES_NO_CANCEL) {
	buttons_type = GTK_BUTTONS_YES_NO;
    }
    
    dialog = gtk_message_dialog_new(parent,
				    GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
				    message_type,
				    buttons_type,
				    "%s", message);
    
    if (SpDialogPart(component).button_type == SP_MB_OK_CANCEL_APPLY) {
	gtk_dialog_add_button(GTK_DIALOG(dialog), SP_GTK_APPLY_ID, GTK_RESPONSE_APPLY);
    } else if (SpDialogPart(component).button_type == SP_MB_YES_NO_CANCEL) {
	gtk_dialog_add_button(GTK_DIALOG(dialog), SP_GTK_CANCEL_ID, GTK_RESPONSE_CANCEL);
    } else if (SpDialogPart(component).button_type == SP_MB_RETRY_CANCEL) {
	gtk_dialog_add_button(GTK_DIALOG(dialog), SP_GTK_APPLY_ID, GTK_RESPONSE_APPLY);
	gtk_dialog_add_button(GTK_DIALOG(dialog), SP_GTK_CANCEL_ID, GTK_RESPONSE_CANCEL);
    }

    //result = gtk_dialog_run(GTK_DIALOG(dialog));
    result = modalLoop(dialog);

    if (result == GTK_RESPONSE_OK) {
	response = SP_DR_OK;
    } else if (result == GTK_RESPONSE_CANCEL) {
	response = SP_DR_CANCEL;
    } else if (result == GTK_RESPONSE_YES) {
	response = SP_DR_YES;
    } else if (result == GTK_RESPONSE_NO) {
	response = SP_DR_NO;
    } else if (result == GTK_RESPONSE_APPLY) {
	response = SP_DR_APPLY;
    }

    gtk_widget_destroy(dialog);
    
    return response;
}
#else
static gint deleteEventCB(GtkWidget *widget, GdkEvent *event, void *data)
{
    mainQuit(widget, SP_FALSE);
    return TRUE;
}

static void messageBoxCB(GtkWidget *widget, spDialogResponse *response)
{
    if (response != NULL) {
	*response = GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(widget), "value"));
	spDebug(30, "messageBoxCB", "*response = %d\n", *response);
    }
    
    mainQuit(widget, SP_TRUE);
    
    spDebug(30, "messageBoxCB", "done\n");
    
    return;
}

static gint keyPressEventCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
    gint flag = FALSE;

    if (event->type == GDK_KEY_PRESS) {
	if (event->key.keyval == GDK_Escape) {
	    mainQuit(widget, SP_FALSE);
	    flag = TRUE;
	}
    }

    return flag;
}

spDialogResponse spPopupMessageBox(spComponent component, const char *message)
{
    spDialogResponse response = SP_DR_NONE;
    char *title;
    GtkWidget *label;
    GtkWidget *vbox;
    GtkWidget *dialog = NULL;
    GtkWidget *button1 = NULL;
    GtkWidget *button2 = NULL;
    GtkWidget *button3 = NULL;

    title = spGetMessageBoxTitle(component);
    
    dialog = gtk_dialog_new();
    vbox = GTK_DIALOG(dialog)->vbox;

    if (!strnone(title)) {
	spDebug(30, "createMessageDialog", "title = %s\n", title);
	gtk_window_set_title(GTK_WINDOW(dialog), title);
    }

    if (!strnone(message)) {
	spDebug(30, "createMessageDialog", "message = %s\n", message);
	
	label = gtk_label_new(message);
	/*gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);*/
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 0);
	gtk_widget_show(label);
    }

    gtk_signal_connect(GTK_OBJECT(dialog), "delete_event",
		       GTK_SIGNAL_FUNC(deleteEventCB), NULL);
    gtk_signal_connect(GTK_OBJECT(dialog), "key_press_event",
		       GTK_SIGNAL_FUNC(keyPressEventCB), NULL);
    
    if (SpDialogPart(component).button_type == SP_MB_OK) {
	button1 = gtk_button_new_with_label(SP_OK_LABEL);
	GTK_WIDGET_SET_FLAGS(button1, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button1, TRUE, TRUE, 2);
	gtk_object_set_data(GTK_OBJECT(button1), "value", (gpointer)SP_DR_OK);
	gtk_signal_connect(GTK_OBJECT(button1), "clicked",
			   GTK_SIGNAL_FUNC(messageBoxCB), &response);
	gtk_widget_grab_default(button1);
	gtk_widget_show(button1);
    } else if (SpDialogPart(component).button_type == SP_MB_OK_CANCEL
	       || SpDialogPart(component).button_type == SP_MB_OK_CANCEL_APPLY) {
	button1 = gtk_button_new_with_label(SP_OK_LABEL);
	GTK_WIDGET_SET_FLAGS(button1, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button1, TRUE, TRUE, 2);
	gtk_object_set_data(GTK_OBJECT(button1), "value", (gpointer)SP_DR_OK);
	gtk_signal_connect(GTK_OBJECT(button1), "clicked",
			   GTK_SIGNAL_FUNC(messageBoxCB), &response);
	gtk_widget_grab_default(button1);
	gtk_widget_show(button1);

	button2 = gtk_button_new_with_label(SP_CANCEL_LABEL);
	GTK_WIDGET_SET_FLAGS(button2, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button2, TRUE, TRUE, 2);
	gtk_object_set_data(GTK_OBJECT(button2), "value", (gpointer)SP_DR_CANCEL);
	gtk_signal_connect(GTK_OBJECT(button2), "clicked",
			   GTK_SIGNAL_FUNC(messageBoxCB), &response);
	gtk_widget_show(button2);
    } else if (SpDialogPart(component).button_type == SP_MB_YES_NO) {
	button1 = gtk_button_new_with_label(SP_YES_LABEL);
	GTK_WIDGET_SET_FLAGS(button1, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button1, TRUE, TRUE, 2);
	gtk_object_set_data(GTK_OBJECT(button1), "value", (gpointer)SP_DR_YES);
	gtk_signal_connect(GTK_OBJECT(button1), "clicked",
			   GTK_SIGNAL_FUNC(messageBoxCB), &response);
	gtk_widget_grab_default(button1);
	gtk_widget_show(button1);

	button2 = gtk_button_new_with_label(SP_NO_LABEL);
	GTK_WIDGET_SET_FLAGS(button2, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button2, TRUE, TRUE, 2);
	gtk_object_set_data(GTK_OBJECT(button2), "value", (gpointer)SP_DR_NO);

	gtk_signal_connect(GTK_OBJECT(button2), "clicked",
			   GTK_SIGNAL_FUNC(messageBoxCB), &response);
	gtk_widget_show(button2);
    } else if (SpDialogPart(component).button_type == SP_MB_YES_NO_CANCEL) {
	button1 = gtk_button_new_with_label(SP_YES_LABEL);
	GTK_WIDGET_SET_FLAGS(button1, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button1, TRUE, TRUE, 2);
	gtk_object_set_data(GTK_OBJECT(button1), "value", (gpointer)SP_DR_YES);
	gtk_signal_connect(GTK_OBJECT(button1), "clicked",
			   GTK_SIGNAL_FUNC(messageBoxCB), &response);
	gtk_widget_grab_default(button1);
	gtk_widget_show(button1);

	button2 = gtk_button_new_with_label(SP_NO_LABEL);
	GTK_WIDGET_SET_FLAGS(button2, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button2, TRUE, TRUE, 2);
	gtk_object_set_data(GTK_OBJECT(button2), "value", (gpointer)SP_DR_NO);
	gtk_signal_connect(GTK_OBJECT(button2), "clicked",
			   GTK_SIGNAL_FUNC(messageBoxCB), &response);
	gtk_widget_show(button2);

	button3 = gtk_button_new_with_label(SP_CANCEL_LABEL);
	GTK_WIDGET_SET_FLAGS(button3, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button3, TRUE, TRUE, 2);
	gtk_object_set_data(GTK_OBJECT(button3), "value", (gpointer)SP_DR_CANCEL);
	gtk_signal_connect(GTK_OBJECT(button3), "clicked",
			   GTK_SIGNAL_FUNC(messageBoxCB), &response);
	gtk_widget_show(button3);
    } else if (SpDialogPart(component).button_type == SP_MB_RETRY_CANCEL) {
	button1 = gtk_button_new_with_label(SP_RETRY_LABEL);
	GTK_WIDGET_SET_FLAGS(button1, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button1, TRUE, TRUE, 2);
	gtk_object_set_data(GTK_OBJECT(button1), "value", (gpointer)SP_DR_RETRY);
	gtk_signal_connect(GTK_OBJECT(button1), "clicked",
			   GTK_SIGNAL_FUNC(messageBoxCB), &response);
	gtk_widget_grab_default(button1);
	gtk_widget_show(button1);

	button2 = gtk_button_new_with_label(SP_CANCEL_LABEL);
	GTK_WIDGET_SET_FLAGS(button2, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), button2, TRUE, TRUE, 2);
	gtk_object_set_data(GTK_OBJECT(button2), "value", (gpointer)SP_DR_CANCEL);
	gtk_signal_connect(GTK_OBJECT(button2), "clicked",
			   GTK_SIGNAL_FUNC(messageBoxCB), &response);
	gtk_widget_show(button2);
    }

    spDebug(30, "createMessageDialog", "before modalLoop\n");
    
    modalLoop(dialog);

    spDebug(30, "createMessageDialog", "done: response = %d\n", response);
    
    return response;
}
#endif

#if GTK_CHECK_VERSION(2,4,0)
char *xspPopupFileSelectionBoxArg(spComponent parent, const char *name,
				  spFileDialogType dialog_type,
				  spArg *args, int num_arg)
{
    GtkWidget *dialog = NULL;
    GtkFileFilter **filters = NULL;
    spComponent component;
    gint result;
    int i;
    int num_filter = 0;
    char *filename = NULL;

    if (dialog_type == SP_FD_TYPE_SAVE) {
	dialog = gtk_file_chooser_dialog_new(SP_SAVE_TITLE, NULL,
					     GTK_FILE_CHOOSER_ACTION_SAVE,
					     SP_GTK_CANCEL_ID, GTK_RESPONSE_CANCEL,
					     SP_GTK_SAVE_ID, GTK_RESPONSE_ACCEPT,
					     NULL);
    } else if (dialog_type == SP_FD_TYPE_OPEN) {
	dialog = gtk_file_chooser_dialog_new(SP_OPEN_TITLE, NULL,
					     GTK_FILE_CHOOSER_ACTION_OPEN,
					     SP_GTK_CANCEL_ID, GTK_RESPONSE_CANCEL,
					     SP_GTK_OPEN_ID, GTK_RESPONSE_ACCEPT,
					     NULL);
    } else if (dialog_type == SP_FD_TYPE_DIR) {
	dialog = gtk_file_chooser_dialog_new(SP_DIR_SELECTION_TITLE, NULL,
					     GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
					     SP_GTK_CANCEL_ID, GTK_RESPONSE_CANCEL,
					     SP_GTK_OPEN_ID, GTK_RESPONSE_ACCEPT,
					     NULL);
    } else {
	return NULL;
    }
    
    component = spCreateComponentArg(SpDialogClass, SpFileDialog,
				     name, NULL, args, num_arg);

    if (dialog_type == SP_FD_TYPE_SAVE) {
	gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(dialog),
						       SpDialogPart(component).overwrite_prompt);
    }
    
    if (SpDialogPart(component).file_filters != NULL) {
	for (i = 0; SpDialogPart(component).file_filters[i] != NULL; i++);
	num_filter = i;
	
	filters = xalloc(num_filter, GtkFileFilter *);
	for (i = 0; i < num_filter; i++) {
	    filters[i] = gtk_file_filter_new();
	    gtk_file_filter_add_pattern(filters[i], SpDialogPart(component).file_filters[i]);
	    
	    if (SpDialogPart(component).file_types != NULL) {
		gtk_file_filter_set_name(filters[i], SpDialogPart(component).file_types[i]);
	    } else {
		gtk_file_filter_set_name(filters[i], SpDialogPart(component).file_filters[i]);
	    }
	    gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filters[i]);
	}

	if (SpDialogPart(component).filter_index != NULL) {
	    gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(dialog),
					filters[*SpDialogPart(component).filter_index]);
	}

    }

    do {
	//result = gtk_dialog_run(GTK_DIALOG(dialog));
	result = modalLoop(dialog);

	if (result == GTK_RESPONSE_ACCEPT) {
	    gchar *gstr;
	
	    gstr = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
	    spDebug(30, "xspPopupFileSelectionBoxArg", "gstr = %s\n", gstr);
	    filename = strclone(gstr);
	    g_free(gstr);
	}
    
	if (filters != NULL) {
	    int selected_filter_index = -1;
	    GtkFileFilter *current_filter;
	
	    current_filter = gtk_file_chooser_get_filter(GTK_FILE_CHOOSER(dialog));
	
	    for (i = 0; i < num_filter; i++) {
		if (result == GTK_RESPONSE_ACCEPT && current_filter == filters[i]) {
		    selected_filter_index = i;
		    if (SpDialogPart(component).filter_index != NULL) {
			*(SpDialogPart(component).filter_index) = i;
		    }
		}
		gtk_file_chooser_remove_filter(GTK_FILE_CHOOSER(dialog), filters[i]);
	    }
	    xfree(filters);

	    if (result == GTK_RESPONSE_ACCEPT
		&& dialog_type == SP_FD_TYPE_SAVE && spGetSuffix(filename) == NULL) {
		char buf[SP_MAX_PATHNAME];

		spStrCopy(buf, sizeof(buf), filename);
		xfree(filename);
	    
		if (spAddCurrentFilterSuffix(component, selected_filter_index, buf, SP_TRUE) == SP_FALSE) {
		    filename = NULL;
		} else {
		    filename = strclone(buf);
		}
	    }
	}
    } while (result == GTK_RESPONSE_ACCEPT && filename == NULL);
    
    spDestroyComponent(component);

    gtk_widget_destroy(dialog);
    
    return filename;
}
#else
static char sp_fd_filename[SP_MAX_PATHNAME] = "";

static void file_selection_ok(GtkWidget *widget,
			      spComponent component)
{
    struct stat status;
    static char buf[SP_MAX_PATHNAME];

    if (component == NULL ||
	SpPrimitiveArch(component).widget == NULL) return;
    
    strcpy(sp_fd_filename,
	   gtk_file_selection_get_filename(GTK_FILE_SELECTION(SpPrimitiveArch(component).widget)));
    
    if (stat(sp_fd_filename, &status) == 0) {
	if (S_ISDIR(status.st_mode)) { /* name is directory name */
	    if (SpDialogPart(component).dialog_type != SP_DIR_SELECTION_DIALOG) {
		SpDialogPart(component).initial_dir = sp_fd_filename;
		strcpy(sp_fd_filename, "");
		return;
	    }
	} else if (SpDialogPart(component).dialog_type == SP_DIR_SELECTION_DIALOG) {
	    spDisplayError(component, NULL, SP_NOT_DIR_MESSAGE);
	    strcpy(sp_fd_filename, "");
	    return;
	} else if (SpDialogPart(component).overwrite_prompt == SP_TRUE) {
	    if (spOverwritePrompt(component, sp_fd_filename) == SP_FALSE) {
		strcpy(sp_fd_filename, "");
		return;
	    }
	}
    } else if (SpDialogPart(component).dialog_type == SP_DIR_SELECTION_DIALOG) {
	/* not found such directory */
	spDisplayError(component, NULL,
		       SP_NOT_FOUND_PATH_MESSAGE);
	strcpy(sp_fd_filename, "");
	return;
    } else if (SpDialogPart(component).file_must_exist == SP_TRUE) {
	spDisplayError(component, NULL,
		       SP_NOT_FOUND_FILE_MESSAGE);
	strcpy(sp_fd_filename, "");
	return;
    } else if (SpDialogPart(component).path_must_exist == SP_TRUE) {
	strcpy(buf, sp_fd_filename);
	spGetDirName(buf);
	
	if (stat(buf, &status) != 0 || !S_ISDIR(status.st_mode)) {
	    /* not found such directory */
	    spDisplayError(component, NULL,
			   SP_NOT_FOUND_PATH_MESSAGE);
	    strcpy(sp_fd_filename, "");
	    return;
	}
    }
    
    mainQuit(widget, SP_TRUE);
    return;
}

static void file_selection_cancel(GtkWidget *widget,
				  spComponent component)
{
    mainQuit(widget, SP_FALSE);
    return;
}

static gint file_selection_delete(GtkWidget *widget, GdkEvent *event,
				  gpointer data)
{
    mainQuit(widget, SP_FALSE);
    return TRUE;
}

static char *sp_default_file_filters[] =
{
    "*",
    NULL,
};

static void setFileMask(spComponent component, int index, spBool complete)
{
    char **file_filters;
    static char **prev_file_filters = NULL;
    static int prev_index = -1;

    if (component == NULL) return;

    if (index < 0) return;
    
    spDebug(10, "setFileMask", "index =  %d\n", index);
    
    if (SpDialogPart(component).filter_index != NULL) {
	*(SpDialogPart(component).filter_index) = index;
    }
    SpPrimitivePart(component).index = index;
    
    if (SpDialogPart(component).file_filters != NULL) {
	file_filters = SpDialogPart(component).file_filters;
    } else {
	file_filters = sp_default_file_filters;
    }

    if (file_filters == prev_file_filters && index == prev_index)
	return;

    if (complete == SP_TRUE) {
	if (SpDialogPart(component).dialog_type == SP_OPEN_FILE_DIALOG) {
#if GTK_CHECK_VERSION(1,2,0)
	    gtk_file_selection_complete(GTK_FILE_SELECTION(SpPrimitiveArch(component).widget), 
					file_filters[index]);
#endif
	} else if (SpDialogPart(component).dialog_type == SP_SAVE_FILE_DIALOG) {
	    char buf[SP_MAX_PATHNAME];
	
	    spStrCopy(buf, sizeof(buf),
		      (char *)gtk_file_selection_get_filename(GTK_FILE_SELECTION(SpPrimitiveArch(component).widget)));
	    if (!strnone(buf)) {
		spReplaceSuffix(buf, file_filters[index]);
		gtk_file_selection_set_filename(GTK_FILE_SELECTION(SpPrimitiveArch(component).widget), buf);
	    }
	}
    }
    
    prev_index = index;
    prev_file_filters = file_filters;
    
    return;
}

static void selectFileTypeCB(spComponent component, void *data)
{
    int index;
    spComponent dialog = (spComponent)data;

    /* get selected index of combo box */
    index = spGetSelectedListIndex(component);

    if (GTK_WIDGET_MAPPED(SpPrimitiveArch(dialog).widget)) {
	setFileMask(dialog, index, SP_TRUE);
    } else {
	setFileMask(dialog, index, SP_FALSE);
    }
    
    return;
}

static void setFileTypes(spComponent component)
{
    int i;
    char **file_types;

    if (SpDialogPart(component).file_types != NULL) {
	file_types = SpDialogPart(component).file_types;
    } else if (SpDialogPart(component).file_filters != NULL) {
	file_types = SpDialogPart(component).file_filters;
    } else {
	file_types = sp_default_file_filters;
    }
    
    if (SpDialogArch(component).file_type_field == NULL) {
	SpObjectPart(component).create_flag = SP_TRUE;
	SpDialogArch(component).file_type_field =
	    spCreateParamField(component, "fileTypeField", 70,
			       SppTitle, SP_FILE_TYPE_LABEL,
			       SppSelectOnly, SP_TRUE,
			       SppEditable, SP_FALSE,
			       SppFieldOffset, 110,
			       SppFieldSize, 180,
			       SppFieldType, SP_FIELD_TYPE_COMBO_BOX,
			       SppFieldStrings, file_types,
			       SppCallbackFunc, selectFileTypeCB,
			       SppCallbackData, component,
			       NULL);
	spAdjustComponentSize(SpParamFieldPart(SpDialogArch(component).file_type_field).box);
    } else {
	while (spDeleteListIndex(SpDialogArch(component).file_type_field, 0) >= 0);
	
	for (i = 0; file_types[i] != NULL; i++) {
	    spAddListItem(SpDialogArch(component).file_type_field, file_types[i]);
	}
    }

    if (SpDialogPart(component).filter_index != NULL) {
	SpPrimitivePart(component).index = MAX(*(SpDialogPart(component).filter_index), 0);
    } else {
	SpPrimitivePart(component).index = MAX(SpPrimitivePart(component).index, 0);
    }
    
    /* select index of list */
    spSelectListIndex(SpDialogArch(component).file_type_field, SpPrimitivePart(component).index);

    /* set file mask */
    setFileMask(component, SpPrimitivePart(component).index, SP_FALSE);

    spDebug(10, "setFileTypes", "done\n");
    
    return;
}

static void createFileDialog(spComponent component, spFileDialogType dialog_type)
{
    GtkWidget *dialog;
    
    if (spIsCreated(component) == SP_FALSE) {
	spDebug(10, "createFileDialog", "create\n");
    
	if (dialog_type == SP_FD_TYPE_SAVE) {
	    dialog = gtk_file_selection_new(SP_SAVE_TITLE);
	    gtk_file_selection_show_fileop_buttons(GTK_FILE_SELECTION(dialog));
	} else if (dialog_type == SP_FD_TYPE_DIR) {
	    dialog = gtk_file_selection_new(SP_DIR_SELECTION_TITLE);
	    gtk_file_selection_show_fileop_buttons(GTK_FILE_SELECTION(dialog));
	} else {
	    dialog = gtk_file_selection_new(SP_OPEN_TITLE);
	    gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(dialog));
	}
	gtk_window_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
	gtk_signal_connect(GTK_OBJECT(dialog), "delete_event",
			   GTK_SIGNAL_FUNC(file_selection_delete), NULL);
	gtk_signal_connect(GTK_OBJECT(dialog), "key_press_event",
			   GTK_SIGNAL_FUNC(keyPressEventCB), NULL);
	SpPrimitiveArch(component).widget = dialog;
	
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(dialog)->ok_button),
			   "clicked", GTK_SIGNAL_FUNC(file_selection_ok),
			   (gpointer)component);
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(dialog)->cancel_button),
			   "clicked", GTK_SIGNAL_FUNC(file_selection_cancel),
			   (gpointer)component);
	spDebug(10, "createFileDialog", "create done\n");
    }

    if (!strnone(SpComponentPart(component).title)) {
	spDebug(10, "createFileDialog", "title = %s\n", SpComponentPart(component).title);
    
	gtk_window_set_title(GTK_WINDOW(SpPrimitiveArch(component).widget),
			     SpComponentPart(component).title);
    }

    return;
}

char *xspPopupFileSelectionBoxArg(spComponent parent, const char *name,
				  spFileDialogType dialog_type,
				  spArg *args, int num_arg)
{
    spComponent component;
    static spComponent sp_open_dialog = NULL;
    static spComponent sp_save_dialog = NULL;
    static spComponent sp_dir_dialog = NULL;

    spDebug(10, "xspPopupFileSelectionBoxArg", "in: num_arg = %d\n", num_arg);
    
    if ((dialog_type == SP_FD_TYPE_SAVE && sp_save_dialog == NULL) ||
	(dialog_type == SP_FD_TYPE_OPEN && sp_open_dialog == NULL) ||
	(dialog_type == SP_FD_TYPE_DIR && sp_dir_dialog == NULL)) {
	component = spCreateComponentArg(SpDialogClass, SpFileDialog,
					 name, NULL, args, num_arg);
	SpObjectPart(component).create_flag = SP_FALSE;
	createFileDialog(component, dialog_type);
	SpObjectPart(component).create_flag = SP_TRUE;
	
	if (dialog_type == SP_FD_TYPE_SAVE) {
	    SpDialogPart(component).dialog_type = SP_SAVE_FILE_DIALOG;
	    sp_save_dialog = component;
	} else if (dialog_type == SP_FD_TYPE_DIR) {
	    SpDialogPart(component).dialog_type = SP_DIR_SELECTION_DIALOG;
	    sp_dir_dialog = component;
	} else {
	    SpDialogPart(component).dialog_type = SP_OPEN_FILE_DIALOG;
	    sp_open_dialog = component;
	}
    } else {
	if (dialog_type == SP_FD_TYPE_SAVE) {
	    component = sp_save_dialog;
	} else if (dialog_type == SP_FD_TYPE_DIR) {
	    component = sp_dir_dialog;
	} else {
	    component = sp_open_dialog;
	}
	spSetParamsArg(component, args, num_arg);
	createFileDialog(component, dialog_type);
    }
    
    setFileTypes(component);
    
    if (!strnone(SpDialogPart(component).initial_file_name)) {
	gtk_file_selection_set_filename(GTK_FILE_SELECTION(SpPrimitiveArch(component).widget), 
					spGetBaseName(SpDialogPart(component).initial_file_name));
    } else {
	gtk_file_selection_set_filename(GTK_FILE_SELECTION(SpPrimitiveArch(component).widget), "");
    }
    
    do {
	strcpy(sp_fd_filename, "");

	modalLoop(SpPrimitiveArch(component).widget);

	if (!strnone(sp_fd_filename)) {
	    if (dialog_type == SP_FD_TYPE_SAVE && spGetSuffix(sp_fd_filename) == NULL) {
		if (spAddCurrentFilterSuffix(component, SpPrimitivePart(component).index,
					     sp_fd_filename, SP_TRUE) == SP_FALSE) {
		    continue;
		}
	    }
	    
	    return strclone(sp_fd_filename);
	} else {
	    break;
	}
    } while (1);
    
    return NULL;
}
#endif

#if GTK_CHECK_VERSION(3,2,0)
char *xspChooseFontArch(spComponent component)
{
    GtkWidget *dialog;
    gint result;
    char *string = NULL;

    dialog = gtk_font_chooser_dialog_new(SP_FONT_SELECTION_TITLE, NULL);

    if (!strnone(SpDialogPart(component).initial_font)) {
	spDebug(10, "createFontDialog", "initial_font = %s\n", SpDialogPart(component).initial_font);
	gtk_font_chooser_set_font(GTK_FONT_CHOOSER(dialog), SpDialogPart(component).initial_font);
    }

    //result = gtk_dialog_run(GTK_DIALOG(dialog));
    result = modalLoop(dialog);

    if (result == GTK_RESPONSE_OK) {
	gchar *gstr;
	gstr = gtk_font_chooser_get_font(GTK_FONT_CHOOSER(dialog));
	string = strclone(gstr);
	g_free(gstr);
    }

    gtk_widget_destroy(dialog);
    
    return string;
}
#elif GTK_CHECK_VERSION(1,2,0)

static char sp_selected_font[SP_MAX_PATHNAME] = "";

#if !GTK_CHECK_VERSION(2,4,0)
static void font_selection_ok (GtkWidget *w,
			       GtkFontSelectionDialog *fs)
{
    gchar *s = gtk_font_selection_dialog_get_font_name(fs);

    if (s != NULL) {
	spDebug(10, "font_selection_ok", "font = %s\n", s);
	spStrCopy(sp_selected_font, SP_MAX_PATHNAME, s);
    
	g_free(s);
    } else {
	strcpy(sp_selected_font, "");
    }
    
    mainQuit(w, SP_TRUE);
    
    return;
}

static void font_selection_cancel(GtkWidget *widget,
				  GtkFontSelectionDialog *fs)
{
    strcpy(sp_selected_font, "");
    mainQuit(widget, SP_FALSE);
    return;
}

static gint font_selection_delete(GtkWidget *widget, GdkEvent *event,
				  gpointer data)
{
    strcpy(sp_selected_font, "");
    mainQuit(widget, SP_FALSE);
    return TRUE;
}
#endif

static GtkWidget *createFontDialog(const char *initial_font)
{
    GtkWidget *dialog;
    
    dialog = gtk_font_selection_dialog_new(SP_FONT_SELECTION_TITLE);
#if GTK_CHECK_VERSION(3,0,0)
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
#else    
    gtk_window_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
#endif

    if (!strnone(initial_font)) {
	spDebug(10, "createFontDialog", "initial_font = %s\n", initial_font);
	gtk_font_selection_dialog_set_font_name(GTK_FONT_SELECTION_DIALOG(dialog),
						initial_font);
    }

#if !GTK_CHECK_VERSION(2,4,0)
    gtk_signal_connect(GTK_OBJECT(dialog), "delete_event",
		       GTK_SIGNAL_FUNC(font_selection_delete), NULL);
    gtk_signal_connect (GTK_OBJECT(GTK_FONT_SELECTION_DIALOG(dialog)->ok_button),
			"clicked", GTK_SIGNAL_FUNC(font_selection_ok),
			GTK_FONT_SELECTION_DIALOG(dialog));
    gtk_signal_connect(GTK_OBJECT(GTK_FONT_SELECTION_DIALOG(dialog)->cancel_button),
		       "clicked", GTK_SIGNAL_FUNC(font_selection_cancel),
		       GTK_FONT_SELECTION_DIALOG(dialog));
#endif
	
    return dialog;
}
     
char *xspChooseFontArch(spComponent component)
{
    gint result;
    static GtkWidget *dialog = NULL;

    if (dialog == NULL) {
	dialog = createFontDialog(SpDialogPart(component).initial_font);
    }
    
    strcpy(sp_selected_font, "");
    
    result = modalLoop(dialog);
    
#if GTK_CHECK_VERSION(2,4,0)
    strcpy(sp_selected_font, "");
    
    if (result == GTK_RESPONSE_OK) {
	gchar *s = gtk_font_selection_dialog_get_font_name(GTK_FONT_SELECTION_DIALOG(dialog));
	if (s != NULL) {
	    spDebug(10, "xspChooseFontArch", "font = %s\n", s);
	    spStrCopy(sp_selected_font, SP_MAX_PATHNAME, s);
	    g_free(s);
	}
    }
    gtk_widget_destroy(dialog);
    dialog = NULL;
#endif

    if (!strnone(sp_selected_font)) {
	return strclone(sp_selected_font);
    }
    
    return NULL;
}
#else
#include <sp/spDialogBoxP.h>

char *xspChooseFontArch(spComponent component)
{
    return xspChooseFontDefault(component);
}
#endif

#if 1
static char sp_selected_color[SP_MAX_PATHNAME] = "";

#if !GTK_CHECK_VERSION(3,4,0)
static GtkColorSelection *spGetColorSelectionGtk(GtkColorSelectionDialog *csd)
{
    GtkColorSelection *colorsel;
    
#if GTK_CHECK_VERSION(3,0,0)
    colorsel = GTK_COLOR_SELECTION(gtk_color_selection_dialog_get_color_selection(csd));
#else
    colorsel = GTK_COLOR_SELECTION(csd->colorsel);
#endif

    return colorsel;
}
#endif

#if !GTK_CHECK_VERSION(2,4,0)
static void updateColorName(gdouble *color)
{
    spPixel pixel;

    if (color == NULL) {
	strcpy(sp_selected_color, "");
    } else {
	pixel = spRGB(spRound(255 * color[0]), spRound(255 * color[1]), spRound(255 * color[2]));
	spGetColorName(pixel, sp_selected_color);
    }

    return;
}

static void color_selection_ok(GtkWidget *w,
			       gpointer data)
{
    GtkWidget *dialog = (GtkWidget *)data;
    GtkColorSelectionDialog *cs;
    GtkColorSelection *colorsel;
    gdouble color[4];

    cs = GTK_COLOR_SELECTION_DIALOG(dialog);
    colorsel = spGetColorSelectionGtk(cs);
    
    gtk_color_selection_get_color(colorsel, color);
    updateColorName(color);
    
    mainQuit(w, SP_TRUE);
    
    return;
}

static void color_selection_cancel(GtkWidget *widget,
				   gpointer data)
{
    strcpy(sp_selected_color, "");
    mainQuit(widget, SP_FALSE);
    return;
}

static gint color_selection_delete(GtkWidget *widget, GdkEvent *event,
				   gpointer data)
{
    strcpy(sp_selected_color, "");
    mainQuit(widget, SP_FALSE);
    return TRUE;
}
#endif

static GtkWidget *createColorDialog(GtkWindow *parent, const char *initial_color)
{
    GtkWidget *dialog;
#if GTK_CHECK_VERSION(3,4,0)
    GtkColorChooser *chooser;
#else
    GtkColorSelection *colorsel;
#endif
    
#if GTK_CHECK_VERSION(3,4,0)
    dialog = gtk_color_chooser_dialog_new(SP_COLOR_SELECTION_TITLE, parent);
#else
    dialog = gtk_color_selection_dialog_new(SP_COLOR_SELECTION_TITLE);
#endif
    
#if GTK_CHECK_VERSION(3,0,0)
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
#else    
    gtk_window_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
#endif

#if GTK_CHECK_VERSION(3,4,0)
    chooser = GTK_COLOR_CHOOSER(dialog);
    gtk_color_chooser_set_use_alpha(chooser, FALSE);
    
    if (!strnone(initial_color)) {
        GdkRGBA rgba;
        
        if (spGetColorValueGtk(initial_color, &rgba)) {
            gtk_color_chooser_set_rgba(chooser, &rgba);
        }
    }
#else /* GTK_CHECK_VERSION(3,4,0) */
    colorsel = spGetColorSelectionGtk(GTK_COLOR_SELECTION_DIALOG(dialog));
    
#if GTK_CHECK_VERSION(2,0,0)
    gtk_color_selection_set_has_opacity_control(colorsel, FALSE);
#else
    gtk_color_selection_set_opacity(colorsel, FALSE);
#endif
#if !GTK_CHECK_VERSION(3,0,0)
    gtk_color_selection_set_update_policy(colorsel, GTK_UPDATE_CONTINUOUS);
    gtk_widget_hide(GTK_COLOR_SELECTION_DIALOG(dialog)->help_button);
#endif

    if (!strnone(initial_color)) {
	GdkColor gdkcolor;
#if !GTK_CHECK_VERSION(2,0,0)
	gdouble color[4];
#endif
        
	spGetColorValueGtk(initial_color, &gdkcolor);
	
#if GTK_CHECK_VERSION(2,0,0)
	gtk_color_selection_set_current_color(colorsel, &gdkcolor);
#else
	color[0] = (double)gdkcolor.red / 65535.0;
	color[1] = (double)gdkcolor.green / 65535.0;
	color[2] = (double)gdkcolor.blue / 65535.0;
	color[3] = 0.0;
	
	gtk_color_selection_set_color(colorsel, color);
#endif
    }
#endif /* !GTK_CHECK_VERSION(3,4,0) */

#if !GTK_CHECK_VERSION(2,4,0)
    gtk_signal_connect(GTK_OBJECT(dialog), "delete_event",
		       GTK_SIGNAL_FUNC(color_selection_delete), NULL);
    gtk_signal_connect(GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(dialog)->ok_button),
		       "clicked", GTK_SIGNAL_FUNC(color_selection_ok),
		       dialog);
    gtk_signal_connect(GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(dialog)->cancel_button),
		       "clicked", GTK_SIGNAL_FUNC(color_selection_cancel),
		       dialog);
#endif
	
    return dialog;
}

char *xspChooseColorArch(spComponent component)
{
    gint result;
    static GtkWidget *dialog = NULL;

    if (dialog == NULL) {
	dialog = createColorDialog(NULL, SpDialogPart(component).initial_color);
    }
    
    strcpy(sp_selected_color, "");

    result = modalLoop(dialog);
    
#if GTK_CHECK_VERSION(2,4,0)
    if (result == GTK_RESPONSE_OK) {
	spPixel pixel = 0L;
#if GTK_CHECK_VERSION(3,4,0)
        GtkColorChooser *chooser;
        GdkRGBA rgba;

        chooser = GTK_COLOR_CHOOSER(dialog);
        gtk_color_chooser_get_rgba(chooser, &rgba);

        pixel = spRGB(spRound(rgba.red * 255.0), spRound(rgba.green * 255.0), spRound(rgba.blue * 255.0));
#else
	GtkColorSelection *colorsel;
	GdkColor gdkcolor;
	
	colorsel = spGetColorSelectionGtk(GTK_COLOR_SELECTION_DIALOG(dialog));
	gtk_color_selection_get_current_color(colorsel, &gdkcolor);
	spDebug(50, "spGetColorValueGtk", "r = %d, g = %d, b = %d\n", gdkcolor.red, gdkcolor.green, gdkcolor.blue);
	pixel = spRGB(gdkcolor.red/256, gdkcolor.green/256, gdkcolor.blue/256);
#endif
	
	spGetColorName(pixel, sp_selected_color);
    } else {
	strcpy(sp_selected_color, "");
    }
    gtk_widget_destroy(dialog);
    dialog = NULL;
#endif

    if (!strnone(sp_selected_color)) {
	return strclone(sp_selected_color);
    }
    
    return NULL;
}
#else
#include <sp/spDialogBoxP.h>

char *xspChooseColorArch(spComponent component)
{
    return xspChooseColorDefault(component);
}
#endif

spBool spDisplayHelpArch(spComponent component, const char *path)
{
    return spDisplayHelpDefault(component, path);
}

spBool spGetPageSetupArch(spComponent component, spPaperSize *paper_size, spPaperOrientation *paper_orientation,
			  double *left_margin_mm, double *right_margin_mm, double *top_margin_mm, double *bottom_margin_mm)
{
    return spGetPageSetupDefault(component, paper_size, paper_orientation,
				 left_margin_mm, right_margin_mm, top_margin_mm, bottom_margin_mm);
}
