/*
 *	spDraw_Xm.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xlocale.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/xpm.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spDrawP.h>
#include <sp/spGraphicsP.h>

extern spTopLevel sp_toplevel;

void spGetDisplayInfo(spComponent component)
{
    if (component == NULL) return;
    
    if (SpPrimitiveArch(component).display == NULL) {
	if (SpPrimitiveArch(component).widget == NULL) {
	    SpPrimitiveArch(component).display = SpTopLevelArch(sp_toplevel).display;
	} else {
	    SpPrimitiveArch(component).display = XtDisplay(SpPrimitiveArch(component).widget);
	}
    }
    if (SpPrimitiveArch(component).screen == NULL) {
	if (SpPrimitiveArch(component).widget == NULL) {
	    SpPrimitiveArch(component).screen = SpTopLevelArch(sp_toplevel).screen;
	} else {
	    SpPrimitiveArch(component).screen = XtScreen(SpPrimitiveArch(component).widget);
	}
    }

    return;
}

spBool spIsAlphaSupported(spComponent component)
{
    return SP_FALSE;
}

spBool spIsDrawable(spComponent component)
{
    if (spIsPrimitive(component) == SP_TRUE
	&& SpPrimitiveArch(component).pixmap != None) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

void *spGetImageHandleArch(spComponent component)
{
    return (void *)SpPrimitiveArch(component).pixmap;
}

spBool spCopyImageArch(spComponent src, spComponent dest,
		       int src_x, int src_y, int width, int height,
		       int dest_x, int dest_y)
{
    spDebug(50, "spCopyImageArch", "src_pixmap = %ld, dest_pixmap = %ld\n",
	    SpPrimitiveArch(src).pixmap, SpPrimitiveArch(dest).pixmap);
    
    XCopyArea(SpPrimitiveArch(dest).display, SpPrimitiveArch(src).pixmap,
	      SpPrimitiveArch(dest).pixmap, SpTopLevelArch(sp_toplevel).fg_gc,
	      src_x, src_y,
	      MIN(SpComponentPart(src).client_width - src_x, width),
	      MIN(SpComponentPart(src).client_height - src_y, height),
	      dest_x, dest_y);

    return SP_TRUE;
}

spBool spStretchImageArch(spComponent src, spComponent dest,
			  int src_x, int src_y, int src_width, int src_height,
			  int dest_x, int dest_y, int dest_width, int dest_height)
{
    return spStretchImageDefault(src, dest, src_x, src_y, src_width, src_height,
				 dest_x, dest_y, dest_width, dest_height);
}

void spFillRectangleArch(spComponent component, spGraphics graphics,
			 int x, int y, int width, int height)
{
    XFillRectangle(SpPrimitiveArch(component).display, SpPrimitiveArch(component).pixmap,
		   (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
		    SpGraphicsArch(graphics).gc),
		   x, y, width, height);

    return;
}

void spDrawRectangleArch(spComponent component, spGraphics graphics,
			 int x, int y, int width, int height)
{
    XDrawRectangle(SpPrimitiveArch(component).display, SpPrimitiveArch(component).pixmap,
		   (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
		    SpGraphicsArch(graphics).gc),
		   x, y, width, height);

    return;
}

void spFillArcArch(spComponent component, spGraphics graphics,
		   int x, int y, int width, int height, double angle1, double angle2)
{
    XFillArc(SpPrimitiveArch(component).display, SpPrimitiveArch(component).pixmap,
	     (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
	      SpGraphicsArch(graphics).gc),
	     x, y, width, height, (int)spRound(angle1 * 64.0), (int)spRound(angle2 * 64.0));

    return;
}

void spDrawArcArch(spComponent component, spGraphics graphics,
		   int x, int y, int width, int height, double angle1, double angle2)
{
    XDrawArc(SpPrimitiveArch(component).display, SpPrimitiveArch(component).pixmap,
	     (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
	      SpGraphicsArch(graphics).gc),
	     x, y, width, height, (int)spRound(angle1 * 64.0), (int)spRound(angle2 * 64.0));

    return;
}

void spDrawLineArch(spComponent component, spGraphics graphics,
		    int x1, int y1, int x2, int y2)
{
    XDrawLine(SpPrimitiveArch(component).display, SpPrimitiveArch(component).pixmap,
	      (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
	       SpGraphicsArch(graphics).gc),
	      x1, y1, x2, y2);
#if 1
    XDrawPoint(SpPrimitiveArch(component).display, SpPrimitiveArch(component).pixmap,
	       (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
		SpGraphicsArch(graphics).gc),
	       x2, y2);
#endif
    
    return;
}

void spDrawPointArch(spComponent component, spGraphics graphics,
		     int x, int y)
{
    XDrawPoint(SpPrimitiveArch(component).display, SpPrimitiveArch(component).pixmap,
	       (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
		SpGraphicsArch(graphics).gc),
	       x, y);
    
    return;
}

void spDrawStringArch(spComponent component, spGraphics graphics,
		      int x, int y, const char *string)
{
    int x_shift = 0;
    spAlignment alignment;
    XRectangle ink, log;
    
    if (SpGraphicsPart(graphics).string_rotation != 0) {
	alignment = SpGraphicsPart(graphics).rotation_failed_alignment;
    } else {
	alignment = SpGraphicsPart(graphics).string_alignment;
    }
    
    if (alignment == SP_ALIGNMENT_CENTER || alignment == SP_ALIGNMENT_END) {
	XmbTextExtents((graphics == NULL ? SpTopLevelArch(sp_toplevel).font_set :
			SpGraphicsArch(graphics).font_set),
		       string, strlen(string), &ink, &log);
	if (alignment == SP_ALIGNMENT_CENTER) {
	    x_shift = ink.width / 2;
	} else {
	    x_shift = ink.width;
	}
    }
    
    XmbDrawString(SpPrimitiveArch(component).display, SpPrimitiveArch(component).pixmap,
		  (graphics == NULL ? SpTopLevelArch(sp_toplevel).font_set :
		   SpGraphicsArch(graphics).font_set),
		  (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
		   SpGraphicsArch(graphics).gc),
		  x - x_shift, y, string, strlen(string));
    
    return;
}

spBool spGetStringExtentArch(spComponent component, spGraphics graphics, const char *string,
			     int *x, int *y, int *width, int *height, int *next_x)
{
    XRectangle ink, log;
    
    XmbTextExtents((graphics == NULL ? SpTopLevelArch(sp_toplevel).font_set :
		    SpGraphicsArch(graphics).font_set),
		   string, strlen(string), &ink, &log);
    
    if (x != NULL) *x = log.x;
    if (y != NULL) *y = log.y;
    if (width != NULL) *width = log.width;
    if (height != NULL) *height = log.height;
    
    if (next_x != NULL) {
	*next_x = XmbTextEscapement((graphics == NULL ? SpTopLevelArch(sp_toplevel).font_set :
				     SpGraphicsArch(graphics).font_set),
				    string, strlen(string));
    }
    
    return SP_TRUE;
}

spBool spCreatePolygonArch(spPolygon polygon)
{
    SpPolygonArch(polygon).points = xalloc(polygon->max_point + 1, XPoint);
    return SP_TRUE;
}

spBool spDestroyPolygonArch(spPolygon polygon)
{
    xfree(SpPolygonArch(polygon).points);
    return SP_TRUE;
}

spBool spStartPolygonPointArch(spPolygon polygon, int x, int y)
{
    SpPolygonArch(polygon).points[0].x = x;
    SpPolygonArch(polygon).points[0].y = y;
    
    return SP_TRUE;
}

spBool spAddPolygonPointArch(spPolygon polygon, int x, int y)
{
    SpPolygonArch(polygon).points[polygon->num_point].x = x;
    SpPolygonArch(polygon).points[polygon->num_point].y = y;
    
    return SP_TRUE;
}

spBool spAddPolygonRelativePointArch(spPolygon polygon, int dx, int dy)
{
    SpPolygonArch(polygon).points[polygon->num_point].x
	= polygon->px + dx;
    SpPolygonArch(polygon).points[polygon->num_point].y
	= polygon->py + dy;
    
    return SP_TRUE;
}

spBool spEndPolygonPointArch(spPolygon polygon)
{
    SpPolygonArch(polygon).points[polygon->num_point].x = SpPolygonArch(polygon).points[0].x;
    SpPolygonArch(polygon).points[polygon->num_point].y = SpPolygonArch(polygon).points[0].y;
    return SP_TRUE;
}

spBool spGetPolygonPointArch(spPolygon polygon, int index, int *x, int *y)
{
    if (x != NULL) *x = SpPolygonArch(polygon).points[index].x;
    if (y != NULL) *y = SpPolygonArch(polygon).points[index].y;
    
    return SP_TRUE;
}

spBool spMovePolygonArch(spPolygon polygon, int dx, int dy)
{
    int i;

    for (i = 0; i < polygon->num_point; i++) {
	SpPolygonArch(polygon).points[i].x += dx;
	SpPolygonArch(polygon).points[i].y += dy;
    }
    
    return SP_TRUE;
}

spBool spFillPolygonArch(spComponent component, spGraphics graphics, spPolygon polygon)
{
    XFillPolygon(SpPrimitiveArch(component).display, SpPrimitiveArch(component).pixmap,
		 (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
		  SpGraphicsArch(graphics).gc),
		 SpPolygonArch(polygon).points, polygon->num_point,
		 Complex, CoordModeOrigin);
    
    return SP_TRUE;
}

spBool spDrawPolygonArch(spComponent component, spGraphics graphics, spPolygon polygon)
{
    XDrawLines(SpPrimitiveArch(component).display, SpPrimitiveArch(component).pixmap,
	       (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
		SpGraphicsArch(graphics).gc),
	       SpPolygonArch(polygon).points, polygon->num_point + 1,
	       CoordModeOrigin);
    
    return SP_TRUE;
}

unsigned long spGetDrawingCapsArch(spComponent component)
{
    return SP_DRAWING_CAPS_COPY_IMAGE
	| SP_DRAWING_CAPS_FILL_RECTANGLE | SP_DRAWING_CAPS_DRAW_RECTANGLE
	| SP_DRAWING_CAPS_FILL_ARC | SP_DRAWING_CAPS_DRAW_ARC
	| SP_DRAWING_CAPS_FILL_POLYGON | SP_DRAWING_CAPS_DRAW_POLYGON
	| SP_DRAWING_CAPS_DRAW_LINE
	| SP_DRAWING_CAPS_STRING_EXTENT | SP_DRAWING_CAPS_STRING_EXTENT_ACCURATE
	| SP_DRAWING_CAPS_DRAW_WIDE_LINE | SP_DRAWING_CAPS_DRAW_STYLED_LINE | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_LINE
	| SP_DRAWING_CAPS_DRAW_WIDE_RECTANGLE | SP_DRAWING_CAPS_DRAW_STYLED_RECTANGLE | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_RECTANGLE
	| SP_DRAWING_CAPS_DRAW_WIDE_POLYGON | SP_DRAWING_CAPS_DRAW_STYLED_POLYGON | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_POLYGON
	| SP_DRAWING_CAPS_DRAW_WIDE_ARC | SP_DRAWING_CAPS_DRAW_STYLED_ARC | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_ARC
	;
}
