#include <sp/spDefs.h>
#include <sp/spBase.h>

#include <sp/spTopLevelP.h>
#include <sp/spDrawP.h>
#include <sp/spDrawingPluginP.h>

static HPEN g_null_pen = NULL;
static HBRUSH g_null_brush = NULL;
static HFONT g_sys_font = NULL;

typedef struct _spPluginInstanceEmf
{
    HDC hdc;
    double xfactor, yfactor;

    double paper_width_mm;
    double paper_height_mm;
    double left_margin_mm;
    double right_margin_mm;
    double top_margin_mm;
    double bottom_margin_mm;
} *spPluginInstanceEmf;

static spBool spInitEmf(const char *lang)
{
    if (g_null_pen == NULL) {
	g_null_pen = GetStockObject(NULL_PEN);
	g_null_brush = GetStockObject(HOLLOW_BRUSH);
	g_sys_font = GetStockObject(DEFAULT_GUI_FONT);
    }
    
    return SP_TRUE;
}

static spBool spFreeEmf(void)
{
    return SP_TRUE;
}

static void *spInitInstanceEmf(const char *lang)
{
    spPluginInstanceEmf pinstance;

    pinstance = xalloc(1, struct _spPluginInstanceEmf);
    memset(pinstance, 0, sizeof(struct _spPluginInstanceEmf));
    pinstance->hdc = NULL;
    pinstance->hdc = NULL;
    pinstance->xfactor = 1.0;
    pinstance->yfactor = 1.0;
    pinstance->paper_width_mm = 0.0;
    pinstance->paper_height_mm = 0.0;
    spGetPaperDimensions(SP_PAPER_SIZE_A4, SP_PAPER_ORIENTATION_PORTRAIT,
			 &pinstance->paper_width_mm, &pinstance->paper_height_mm);
    pinstance->left_margin_mm = 20.0;
    pinstance->right_margin_mm = 20.0;
    pinstance->top_margin_mm = 20.0;
    pinstance->bottom_margin_mm = 20.0;
    
    return (void *)pinstance;
}

static spBool spCloseEnhMetaFile(spPluginInstanceEmf pinstance)
{
    HENHMETAFILE hmeta;
    spBool flag = SP_FALSE;

    if (pinstance->hdc != NULL) {
	if ((hmeta = CloseEnhMetaFile(pinstance->hdc)) != NULL) {
	    DeleteEnhMetaFile(hmeta);
	    flag = SP_TRUE;
	}
	pinstance->hdc = NULL;
    }
    
    return flag;
}

static spBool spFreeInstanceEmf(void *instance)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;

    if (pinstance->hdc != NULL) {
	spCloseEnhMetaFile(pinstance);
    }
    
    xfree(instance);
    return SP_TRUE;
}

static spBool spCreateComponentEmf(void *instance, void *parent_arch, spArg *args, int num_arg)
{
    return SP_TRUE;
}

static spBool spDestroyComponentEmf(void *instance)
{
    return SP_TRUE;
}

static spBool spOpenTargetFileEmf(void *instance, const char *filename)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;

#if 0
    {
	RECT rect;
	rect.left = 0;
	rect.top = 0;
	rect.right = (LONG)spRound(100.0 * (pinstance->paper_width_mm - pinstance->left_margin_mm - pinstance->right_margin_mm)
				   / pinstance->xfactor);
	rect.bottom = (LONG)spRound(100.0 * (pinstance->paper_height_mm - pinstance->top_margin_mm - pinstance->bottom_margin_mm)
				    / pinstance->yfactor);
	pinstance->hdc = CreateEnhMetaFile(NULL, filename, &rect, NULL);
    }
#else
	pinstance->hdc = CreateEnhMetaFile(NULL, filename, NULL, NULL);
#endif

    return SP_TRUE;
}

    
static spBool spCloseTargetEmf(void *instance)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;

    return spCloseEnhMetaFile(pinstance);
}

spBool spSetPageInfoEmf(void *instance, spPaperSize paper_size, spPaperOrientation paper_orientation,
	double left_margin_mm, double right_margin_mm, double top_margin_mm, double bottom_margin_mm)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;

    pinstance->left_margin_mm = left_margin_mm;
    pinstance->right_margin_mm = right_margin_mm;
    pinstance->top_margin_mm = top_margin_mm;
    pinstance->bottom_margin_mm = bottom_margin_mm;
    
    return spGetPaperDimensions(paper_size, paper_orientation,
				&pinstance->paper_width_mm, &pinstance->paper_height_mm);
}


spBool spGetPageInfoEmf(void *instance, spPaperSize *paper_size, spPaperOrientation *paper_orientation,
        double *left_margin_mm, double *right_margin_mm, double *top_margin_mm, double *bottom_margin_mm)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;

    *left_margin_mm = pinstance->left_margin_mm;
    *right_margin_mm = pinstance->right_margin_mm;
    *top_margin_mm = pinstance->top_margin_mm;
    *bottom_margin_mm = pinstance->bottom_margin_mm;
    
    return spGetPaperFromDimensions(pinstance->paper_width_mm, pinstance->paper_height_mm,
				    paper_size, paper_orientation);
}

unsigned long spGetDrawingCapsEmf(void *instance)
{
    return spGetDrawingCapsWin();
}

unsigned long spGetGraphicsModeCapsEmf(void *instance)
{
    return SP_GRAPHICS_MODE_CAPS_XOR | SP_GRAPHICS_MODE_CAPS_INVERT
	| SP_GRAPHICS_MODE_CAPS_AND | SP_GRAPHICS_MODE_CAPS_OR;
}

static spBool spIsDrawableEmf(void *instance)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;

    if (pinstance->hdc != NULL) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static void convertSizeMMToDot(HDC hdc, double paper_width_mm, double paper_height_mm,
			       double left_margin_mm, double right_margin_mm, double top_margin_mm, double bottom_margin_mm,
			       int *left_margin_dot, int *top_margin_dot, int *client_width_dot, int *client_height_dot)
{
    int x_offset_dot, y_offset_dot;
    double xdpi, ydpi;
    double x_dp_mm, y_dp_mm;
    
    x_offset_dot = 0;
    y_offset_dot = 0;
    spDebug(50, "convertSizeMMToDot", "x_offset_dot = %d, y_offset_dot = %d\n",
	    x_offset_dot, y_offset_dot);

    xdpi = (double)GetDeviceCaps(hdc, LOGPIXELSX);
    ydpi = (double)GetDeviceCaps(hdc, LOGPIXELSY);
    spDebug(50, "convertSizeMMToDot", "xdpi = %f, ydpi = %f\n", xdpi, ydpi);
    
    x_dp_mm = xdpi / 25.4;
    y_dp_mm = ydpi / 25.4;
    spDebug(50, "convertSizeMMToDot", "x_dp_mm = %f, y_dp_mm = %f\n", x_dp_mm, y_dp_mm);

    if (client_width_dot != NULL) {
	*client_width_dot = (int)spRound((paper_width_mm - left_margin_mm - right_margin_mm) * x_dp_mm);
	spDebug(50, "convertSizeMMToDot", "client_width_dot = %d\n", *client_width_dot);
    }
    if (client_height_dot != NULL) {
	*client_height_dot = (int)spRound((paper_height_mm - top_margin_mm - bottom_margin_mm) * y_dp_mm);
	spDebug(50, "convertSizeMMToDot", "client_height_dot = %d\n", *client_height_dot);
    }
    
    if (left_margin_dot != NULL) {
	*left_margin_dot = MAX((int)spRound(left_margin_mm * x_dp_mm) - x_offset_dot, 0);
	spDebug(50, "convertSizeMMToDot", "left_margin_dot = %d\n", *left_margin_dot);
    }
    if (top_margin_dot != NULL) {
	*top_margin_dot = MAX((int)spRound(top_margin_mm * y_dp_mm) - y_offset_dot, 0);
	spDebug(50, "convertSizeMMToDot", "top_margin_dot = %d\n", *top_margin_dot);
    }

    return;
}

static spBool spGetClientSizeEmf(void *instance, int *width, int *height)
{
    int width_dot, height_dot;
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;

    convertSizeMMToDot(pinstance->hdc, pinstance->paper_width_mm, pinstance->paper_height_mm,
		       pinstance->left_margin_mm, pinstance->right_margin_mm,
		       pinstance->top_margin_mm, pinstance->bottom_margin_mm,
		       NULL, NULL, &width_dot, &height_dot);
    spDebug(50, "spGetClientSizeEmf", "width_dot = %d, height_dot = %d\n", width_dot, height_dot);
    *width = (int)((double)width_dot / pinstance->xfactor);
    *height = (int)((double)height_dot / pinstance->yfactor);
    spDebug(50, "spGetClientSizeEmf", "width = %d, height = %d\n", *width, *height);

    return SP_TRUE;
}

static spBool spGetPhysicalSizeEmf(void *instance, double *width_mm, double *height_mm)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;

    *width_mm = pinstance->paper_width_mm - pinstance->left_margin_mm - pinstance->right_margin_mm;
    *height_mm = pinstance->paper_height_mm - pinstance->top_margin_mm - pinstance->bottom_margin_mm;
    
    return SP_TRUE;
}

static spBool spSetScalingFactorEmf(void *instance, double xfactor, double yfactor)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;
    
    pinstance->xfactor = xfactor;
    pinstance->yfactor = yfactor;
    
    return SP_TRUE;
}
     
static spBool transformCoordinate(spPluginInstanceEmf pinstance, double xfactor, double yfactor)
{
    int dev_width, dev_height;
    SIZE window_size;
    SIZE viewport_size;

#if 1
    dev_width = GetDeviceCaps(pinstance->hdc, HORZRES);
    dev_height = GetDeviceCaps(pinstance->hdc, VERTRES);
#else
    convertSizeMMToDot(pinstance->hdc, pinstance->paper_width_mm, pinstance->paper_height_mm,
		       pinstance->left_margin_mm, pinstance->right_margin_mm,
		       pinstance->top_margin_mm, pinstance->bottom_margin_mm,
		       NULL, NULL, &dev_width, &dev_height);
#endif
    spDebug(50, "transformCoordinate", "dev_width = %d, dev_height = %d, xfactor = %f, yfactor = %f\n",
	    dev_width, dev_height, xfactor, yfactor);

    if (SetMapMode(pinstance->hdc, MM_ANISOTROPIC)) {
	if (SetWindowExtEx(pinstance->hdc,
			   (int)spRound((double)dev_width / xfactor),
			   (int)spRound((double)dev_height / yfactor), &window_size)) {
	    spDebug(50, "transformCoordinate", "win width = %ld, win height = %ld\n",
		    window_size.cx, window_size.cy);
	    if (SetViewportExtEx(pinstance->hdc,
				 dev_width, dev_height, &viewport_size)) {
		spDebug(50, "transformCoordinate", "viewport width = %ld, viewport height = %ld\n",
			viewport_size.cx, viewport_size.cy);
		return SP_TRUE;
	    }
	}
    }
    
    return SP_FALSE;
}
     
static spBool spStartDrawingEmf(void *instance, const char *job_name)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;

    transformCoordinate(pinstance, pinstance->xfactor, pinstance->yfactor);
    
    return SP_TRUE;
}

static spBool spEndDrawingEmf(void *instance)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;
    
    return SP_TRUE;
}

static spBool spCopyImageEmf(void *instance, void *src_image_arch,
				 int src_x, int src_y, int src_width, int src_height,
				 int dest_x, int dest_y, int dest_width, int dest_height)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;

    if (StretchBlt(pinstance->hdc, dest_x, dest_y, dest_width, dest_height,
		   (HDC)src_image_arch, src_x, src_y, src_width, src_height, SRCCOPY)) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

void spFillRectangleEmf(void *instance, spGraphics graphics,
			    int x, int y, int width, int height)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;

    spDrawRectangleWin(pinstance->hdc, graphics, SP_GRAPHICS_BRUSH,
		       g_null_pen, g_null_brush, x, y, width, height);

    return;
}

void spDrawRectangleEmf(void *instance, spGraphics graphics,
			    int x, int y, int width, int height)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;

    spDrawRectangleWin(pinstance->hdc, graphics, SP_GRAPHICS_PEN,
		       g_null_pen, g_null_brush, x, y, width, height);

    return;
}

void spFillArcEmf(void *instance, spGraphics graphics,
		      int x, int y, int width, int height, double angle1, double angle2)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;
    
    spDrawArcWin(pinstance->hdc, graphics, SP_GRAPHICS_BRUSH,
		 g_null_pen, g_null_brush,
		 x, y, width, height, angle1, angle2);
    
    return;
}
     
void spDrawArcEmf(void *instance, spGraphics graphics,
		      int x, int y, int width, int height, double angle1, double angle2)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;
    
    spDrawArcWin(pinstance->hdc, graphics, SP_GRAPHICS_PEN,
		 g_null_pen, g_null_brush,
		 x, y, width, height, angle1, angle2);

    return;
}

void spDrawLineEmf(void *instance, spGraphics graphics,
		       int x1, int y1, int x2, int y2)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;
    
    spDrawLineWin(pinstance->hdc, graphics, x1, y1, x2, y2);

    return;
}

void spDrawPointEmf(void *instance, spGraphics graphics,
			int x, int y)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;

    spDrawPointWin(pinstance->hdc, graphics, x, y);

    return;
}

void spDrawStringEmf(void *instance, spGraphics graphics,
			 int x, int y, const char *string)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;
    
    spDrawStringWin(pinstance->hdc, graphics, x, y, string);
}

spBool spGetStringExtentEmf(void *instance, spGraphics graphics, const char *string,
				int *x, int *y, int *width, int *height, int *next_x)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;
    
    return spGetStringExtentWin(pinstance->hdc, graphics, string, x, y, width, height, next_x);
}

spBool spFillPolygonEmf(void *instance, spGraphics graphics, void *polygon_instance)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;
    
    return spDrawPolygonWin(pinstance->hdc, graphics, SP_GRAPHICS_BRUSH,
			    g_null_pen, g_null_brush, (spPolygon)polygon_instance);
}

spBool spDrawPolygonEmf(void *instance, spGraphics graphics, void *polygon_instance)
{
    spPluginInstanceEmf pinstance = (spPluginInstanceEmf)instance;
    spPolygon polygon = (spPolygon)polygon_instance;
    
    return spDrawPolygonWin(pinstance->hdc, graphics, SP_GRAPHICS_PEN,
			    g_null_pen, g_null_brush, (spPolygon)polygon_instance);
}

static char *sp_drawing_plugin_emf_type_list[] = {
    "EMF",
    NULL,
};

static char *sp_drawing_plugin_emf_desc_list[] = {
    "Enhanced Meta File",
    NULL,
};

static char *sp_drawing_plugin_emf_filter_list[] = {
    "*.emf",
    NULL,
};

static spDrawingPluginRec sp_drawing_plugin_emf = {
    NULL,
    NULL,

    SP_PLUGIN_DRAWING,
    SP_DRAWING_PLUGIN_EMF,
    1,
    SP_PLUGIN_PRIORITY_MIDDLE,
    0L,		/* caps */
    spInitEmf,
    spFreeEmf,
    "Enhanced Meta File",
    "Enhanced Meta File Plugin Version 0.1\nCopyright(C) Hideki Banno",
    
    spInitInstanceEmf,
    spFreeInstanceEmf,
    NULL,
    NULL,

    SP_DRAWING_DEVICE_FILE,

    NULL,
    NULL,
    NULL,
    NULL,

    NULL,
    spCreateComponentEmf,
    spDestroyComponentEmf,
    NULL,
    NULL,

    NULL,
    
    spSetPageInfoEmf,
    spGetPageInfoEmf,
    NULL,
    NULL,
    NULL,

    sp_drawing_plugin_emf_type_list,
    sp_drawing_plugin_emf_desc_list,
    sp_drawing_plugin_emf_filter_list,
    NULL,
    NULL,
    
    spOpenTargetFileEmf,
    NULL,

    spCloseTargetEmf,
    
    spIsDrawableEmf,

    spGetDrawingCapsEmf,
    spGetGraphicsModeCapsEmf,
    NULL,
    spGetClientSizeEmf,
    spGetPhysicalSizeEmf,
    spSetScalingFactorEmf,

    spStartDrawingEmf,
    spEndDrawingEmf,
    NULL,

    NULL,
    spCopyImageEmf,

    spFillRectangleEmf,
    spDrawRectangleEmf,
    spFillArcEmf,
    spDrawArcEmf,
    spDrawLineEmf,
    spDrawPointEmf,
    spDrawStringEmf,
    spGetStringExtentEmf,

    NULL,
    NULL,
    NULL,
    
    NULL,
    spFillPolygonEmf,
    spDrawPolygonEmf,
};

spBool spAddEmfPlugin(void)
{
    return spAddPluginRecList((spPluginRec *)&sp_drawing_plugin_emf, SP_DRAWING_PLUGIN_EMF);
}

spBool spRemoveEmfPlugin(void)
{
    return spRemovePluginRecList((spPluginRec *)&sp_drawing_plugin_emf);
}
