/*
 *	psdrawing.c
 */

#include <sp/spDefs.h>

#include <sp/spPSFile.h>
#include <sp/spDrawP.h>
#include <sp/spDialogBoxP.h>
#include <sp/spDrawingPluginP.h>

typedef struct _spPluginInstancePS
{
    spPSFile *psfile;
    unsigned long options;

    char lang[SP_MAX_LINE];
    double paper_width_mm, paper_height_mm;
    double width_mm, height_mm;
    double left_margin_mm;
    double bottom_margin_mm;

    spLineType current_line_type;
    int current_line_width;
    spPixel current_fg_pixel;
    /*int current_font_size;*/
    char current_font[SP_MAX_LINE];
    spAlignment current_string_alignment;
    int current_string_rotation;

    char temp_file[SP_MAX_PATHNAME];
    char *printer_name;
    int start_page;
    int end_page;
} *spPluginInstancePS;

static spBool spInitPS(const char *lang)
{
    return SP_TRUE;
}

static spBool spFreePS(void)
{
    return SP_TRUE;
}

static void initCurrentStatusPS(spPluginInstancePS pinstance)
{
    pinstance->current_line_type = SP_LINE_SOLID;
    pinstance->current_line_width = 1;
    pinstance->current_fg_pixel = 0L;
    /*pinstance->current_font_size = SP_PS_DEFAULT_FONT_SIZE;*/
    spStrCopy(pinstance->current_font, sizeof(pinstance->current_font), SP_PS_DEFAULT_FONT);
    pinstance->current_string_alignment = SP_ALIGNMENT_BEGINNING;
    pinstance->current_string_rotation = 0;
    
    return;
}

static void *spInitInstancePS(const char *lang)
{
    spPluginInstancePS pinstance;

    pinstance = xalloc(1, struct _spPluginInstancePS);
    memset(pinstance, 0, sizeof(struct _spPluginInstancePS));
    pinstance->psfile = NULL;
    pinstance->options = SP_PS_SCREEN_COORDINATE;
    
    spStrCopy(pinstance->lang, sizeof(pinstance->lang), lang);
    spDebug(50, "spInitInstancePS", "lang = %s\n", pinstance->lang);
    pinstance->left_margin_mm = 20.0;
    pinstance->bottom_margin_mm = 20.0;

    spGetPaperDimensions(SP_PAPER_SIZE_A4, SP_PAPER_ORIENTATION_PORTRAIT,
			 &pinstance->paper_width_mm, &pinstance->paper_height_mm);
    
    pinstance->width_mm = pinstance->paper_width_mm - pinstance->left_margin_mm * 2;
    pinstance->height_mm = pinstance->paper_height_mm - pinstance->bottom_margin_mm * 2;
    spDebug(50, "spInitInstancePS", "paper size = (%f, %f), margin = (%f, %f)\n",
            pinstance->paper_width_mm, pinstance->paper_height_mm, pinstance->left_margin_mm, pinstance->bottom_margin_mm);

    initCurrentStatusPS(pinstance);

    pinstance->start_page = 0;
    pinstance->end_page = 0;
    
    return (void *)pinstance;
}

static spBool spFreeInstancePS(void *instance)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    if (pinstance->printer_name != NULL) xfree(pinstance->printer_name);

    xfree(instance);
    
    return SP_TRUE;
}

static spBool spCreateComponentPS(void *instance, void *parent_arch, spArg *args, int num_arg)
{
    return SP_TRUE;
}

static spBool spDestroyComponentPS(void *instance)
{
    return SP_TRUE;
}

spBool spSetPageInfoPS(void *instance, spPaperSize paper_size, spPaperOrientation paper_orientation,
	double left_margin_mm, double right_margin_mm, double top_margin_mm, double bottom_margin_mm)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;
    
    if (spGetPaperDimensions(paper_size, paper_orientation,
			     &pinstance->paper_width_mm, &pinstance->paper_height_mm) == SP_TRUE) {
	pinstance->width_mm = pinstance->paper_width_mm - left_margin_mm - right_margin_mm;
	pinstance->left_margin_mm = left_margin_mm;

	pinstance->height_mm = pinstance->paper_height_mm - top_margin_mm - bottom_margin_mm;
	pinstance->bottom_margin_mm = bottom_margin_mm;

	if (spIsPaperOrientationHorizontal(paper_orientation)) {
	    pinstance->options |= SP_PS_LANDSCAPE;
	} else {
	    pinstance->options &= ~SP_PS_LANDSCAPE;
	}
	
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spGetPageInfoPS(void *instance, spPaperSize *paper_size, spPaperOrientation *paper_orientation,
        double *left_margin_mm, double *right_margin_mm, double *top_margin_mm, double *bottom_margin_mm)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    *left_margin_mm = pinstance->left_margin_mm;
    *right_margin_mm = pinstance->paper_width_mm - pinstance->left_margin_mm - pinstance->width_mm;
    *top_margin_mm = pinstance->paper_height_mm - pinstance->bottom_margin_mm - pinstance->height_mm;
    *bottom_margin_mm = pinstance->bottom_margin_mm;
    
    return spGetPaperFromDimensions(pinstance->paper_width_mm, pinstance->paper_height_mm,
				    paper_size, paper_orientation);
}

spBool spGetPageRangePS(void *instance, int *start_page, int *end_page)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    *start_page = pinstance->start_page;
    *end_page = pinstance->end_page;

    return SP_TRUE;
}

static spBool spSetFileTypePS(void *instance, int index)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    if (index == 1) {
	pinstance->options |= SP_PS_EPS;
    } else {
	pinstance->options &= ~SP_PS_EPS;
    }
    
    return SP_TRUE;
}

static spBool spOpenTargetFilePS(void *instance, const char *filename)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    spDebug(50, "spOpenTargetFilePS", "filename = %s, options = %lx, size = (%f, %f)\n",
            filename, pinstance->options, pinstance->width_mm, pinstance->height_mm);
    
    if ((pinstance->psfile = spPSOpenFile(filename, pinstance->width_mm,
					  pinstance->height_mm, pinstance->options)) != NULL) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool spCloseTargetFilePS(void *instance)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    if (pinstance->psfile != NULL) {
	spPSCloseFile(pinstance->psfile);
        initCurrentStatusPS(pinstance);
    }
    pinstance->psfile = NULL;
    
    return SP_TRUE;
}

static spBool spIsDrawablePS(void *instance)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    if (pinstance->psfile != NULL) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

unsigned long spGetDrawingCapsPS(void *instance)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;
    unsigned long caps = SP_DRAWING_CAPS_FILL_RECTANGLE | SP_DRAWING_CAPS_DRAW_RECTANGLE
	| SP_DRAWING_CAPS_FILL_ARC | SP_DRAWING_CAPS_DRAW_ARC
	| SP_DRAWING_CAPS_FILL_POLYGON | SP_DRAWING_CAPS_DRAW_POLYGON
	| SP_DRAWING_CAPS_DRAW_LINE
	| SP_DRAWING_CAPS_STRING_EXTENT 
	| SP_DRAWING_CAPS_STRING_ROTATION
	| SP_DRAWING_CAPS_DRAW_WIDE_LINE | SP_DRAWING_CAPS_DRAW_STYLED_LINE | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_LINE
	| SP_DRAWING_CAPS_DRAW_WIDE_RECTANGLE | SP_DRAWING_CAPS_DRAW_STYLED_RECTANGLE | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_RECTANGLE
	| SP_DRAWING_CAPS_DRAW_WIDE_POLYGON | SP_DRAWING_CAPS_DRAW_STYLED_POLYGON | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_POLYGON
	| SP_DRAWING_CAPS_DRAW_WIDE_ARC | SP_DRAWING_CAPS_DRAW_STYLED_ARC | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_ARC
	;

    if (!(pinstance->options & SP_PS_EPS)) {
        caps |= SP_DRAWING_CAPS_MULTIPLE_PAGES;
    }
    
    return caps;
}

unsigned long spGetGraphicsModeCapsPS(void *instance)
{
    return 0L;
}

static spBool spGetClientSizePS(void *instance, int *width, int *height)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    if (pinstance->psfile == NULL) return SP_FALSE;
    
    return spPSGetClientSize(pinstance->psfile, width, height);
}

static spBool spGetPhysicalSizePS(void *instance, double *width_mm, double *height_mm)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    if (pinstance->psfile == NULL) return SP_FALSE;

    return spPSGetPhisicalSize(pinstance->psfile, width_mm, height_mm);
}

static spBool spSetScalingFactorPS(void *instance, double xfactor, double yfactor)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    if (pinstance->psfile == NULL) return SP_FALSE;
    
    return spPSSetScalingFactor(pinstance->psfile, xfactor, yfactor);
}
     
static spBool spStartDrawingPS(void *instance, const char *job_name)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    spPSSetLang(pinstance->psfile, pinstance->lang);
    spPSSetMargin(pinstance->psfile, pinstance->left_margin_mm, pinstance->bottom_margin_mm);
    
    return spPSStartDrawing(pinstance->psfile);
}

static spBool spEndDrawingPS(void *instance)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    return spPSEndDrawing(pinstance->psfile);
}

static spBool spForwardPagePS(void *instance)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    return spPSForwardPage(pinstance->psfile);
}

static void getRGBColor(spPixel pixel, double *r, double *g, double *b)
{
    *r = (double)spGetRValue(pixel) / 255.0;
    *g = (double)spGetGValue(pixel) / 255.0;
    *b = (double)spGetBValue(pixel) / 255.0;
}

static void updateGraphicParams(spPluginInstancePS pinstance, spGraphics graphics)
{
    spLineType line_type;
    int line_width;
    spPixel fg_pixel;
    char *font_name;
    spAlignment string_alignment;
    int string_rotation;
    double r, g, b;
    
    spGetGraphicsParams(graphics,
			SppLineType, &line_type,
			SppLineWidth, &line_width,
			SppForegroundPixel, &fg_pixel,
			SppStringAlignment, &string_alignment,
			SppStringRotation, &string_rotation,
			SppFontName, &font_name,
			NULL);
    spDebug(50, "updateGraphicParams",
	    "line_type = %d, line_width = %d, fg_pixel = %lx, string_alignment = %d, string_rotation = %d\n",
	    line_type, line_width, fg_pixel, string_alignment, string_rotation);
    
    if (line_width != pinstance->current_line_width) {
	spPSSetLineWidth(pinstance->psfile, line_width);
	pinstance->current_line_width = line_width;
    }
    if (line_type != pinstance->current_line_type) {
	spPSSetLineType(pinstance->psfile, (spPSLineType)line_type);
	pinstance->current_line_type = line_type;
    }
    if (fg_pixel != pinstance->current_fg_pixel) {
	getRGBColor(fg_pixel, &r, &g, &b);
	spPSSetRGBColor(pinstance->psfile, r, g, b);
	pinstance->current_fg_pixel = fg_pixel;
    }
#if 1
    if (font_name != NULL
	&& !streq(font_name, pinstance->current_font)) {
	spDebug(50, "updateGraphicParams",
		"font_name = %s, current_font = %s\n", font_name, pinstance->current_font);
	spPSSetFont(pinstance->psfile, font_name);
	spStrCopy(pinstance->current_font, sizeof(pinstance->current_font), font_name);
    }
#endif
    if (string_alignment != pinstance->current_string_alignment) {
	spPSSetStringAlignment(pinstance->psfile, string_alignment);
	pinstance->current_string_alignment = string_alignment;
    }
    if (string_rotation != pinstance->current_string_rotation) {
	spPSSetStringRotation(pinstance->psfile, string_rotation);
	pinstance->current_string_rotation = string_rotation;
    }
    
    
    return;
}

void spFillRectanglePS(void *instance, spGraphics graphics,
			    int x, int y, int width, int height)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    updateGraphicParams(pinstance, graphics);
    spPSFillRectangle(pinstance->psfile, x, y, width, height);
    
    return;
}

void spDrawRectanglePS(void *instance, spGraphics graphics,
			    int x, int y, int width, int height)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    updateGraphicParams(pinstance, graphics);
    spPSDrawRectangle(pinstance->psfile, x, y, width, height);
    
    return;
}

void spFillArcPS(void *instance, spGraphics graphics,
		      int x, int y, int width, int height, double angle1, double angle2)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;
    
    updateGraphicParams(pinstance, graphics);
    spPSFillArc(pinstance->psfile, x, y, width, height, angle1, angle2);
    
    return;
}
     
void spDrawArcPS(void *instance, spGraphics graphics,
		      int x, int y, int width, int height, double angle1, double angle2)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;
    
    updateGraphicParams(pinstance, graphics);
    spPSDrawArc(pinstance->psfile, x, y, width, height, angle1, angle2);
    
    return;
}

void spDrawLinePS(void *instance, spGraphics graphics,
		       int x1, int y1, int x2, int y2)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;
    
    updateGraphicParams(pinstance, graphics);
    spPSDrawLine(pinstance->psfile, x1, y1, x2, y2);
    
    return;
}

void spDrawPointPS(void *instance, spGraphics graphics,	int x, int y)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    updateGraphicParams(pinstance, graphics);
    spPSFillRectangle(pinstance->psfile, x, y, 1, 1);

    return;
}

void spDrawStringPS(void *instance, spGraphics graphics,
			 int x, int y, const char *string)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;
    
    updateGraphicParams(pinstance, graphics);
    spPSDrawString(pinstance->psfile, x, y, string);
    
    return;
}

typedef struct _spPolygonInstancePS {
    spPSPolygonId polygon_id;
    int dx;
    int dy;
} *spPolygonInstancePS;

void *spInitPolygonInstancePS(spPolygon polygon)
{
    spPolygonInstancePS polygon_instance;

    polygon_instance = xalloc(1, struct _spPolygonInstancePS);
    polygon_instance->polygon_id = SP_PS_POLYGON_NONE;
    polygon_instance->dx = 0;
    polygon_instance->dy = 0;

    spDebug(80, "spInitPolygonInstancePS", "done\n");
    
    return polygon_instance;
}

spBool spFreePolygonInstancePS(void *instance)
{
    xfree(instance);
    
    spDebug(80, "spFreePolygonInstancePS", "done\n");
    
    return SP_TRUE;
}

spBool spPolygonEventPS(void *instance, void *polygon_instance, spPolygonEventType event_type, int x, int y)
{
    spBool flag = SP_FALSE;
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;
    spPolygonInstancePS polygon_instance_ps = (spPolygonInstancePS)polygon_instance;

    spDebug(50, "spPolygonEventPS", "event_type = %d, x = %d, y = %d\n", event_type, x, y);
    
    switch (event_type) {
      case SP_POLYGON_EVENT_START_POINT:
	flag = spPSStartPolygonPoint(pinstance->psfile, x, y);
	break;
      case SP_POLYGON_EVENT_ADD_POINT:
	flag = spPSAddPolygonPoint(pinstance->psfile, x, y);
	break;
      case SP_POLYGON_EVENT_END_POINT:
	if ((polygon_instance_ps->polygon_id = spPSEndPolygonPoint(pinstance->psfile)) != SP_PS_POLYGON_NONE) {
	    flag = SP_TRUE;
	}
	spDebug(50, "spPolygonEventPS", "END_POINT: polygon_id = %ld\n", polygon_instance_ps->polygon_id);
	break;
      case SP_POLYGON_EVENT_MOVE:
	polygon_instance_ps->dx = x;
	polygon_instance_ps->dy = y;
	spDebug(50, "spPolygonEventPS", "MOVE: dx = %d, dy = %d\n", polygon_instance_ps->dx, polygon_instance_ps->dy);
	flag = SP_TRUE;
	break;
      default:
	break;
    }

    spDebug(50, "spPolygonEventPS", "done: flag = %d\n", flag);
    
    return flag;
}

spBool spFillPolygonPS(void *instance, spGraphics graphics, void *polygon_instance)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;
    spPolygonInstancePS polygon_instance_ps = (spPolygonInstancePS)polygon_instance;
    
    spDebug(50, "spFillPolygonPS", "dx = %d, dy = %d\n", polygon_instance_ps->dx, polygon_instance_ps->dy);
    
    updateGraphicParams(pinstance, graphics);
    return spPSFillPolygon(pinstance->psfile, polygon_instance_ps->polygon_id,
			   polygon_instance_ps->dx, polygon_instance_ps->dy);
}

spBool spDrawPolygonPS(void *instance, spGraphics graphics, void *polygon_instance)
{
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;
    spPolygonInstancePS polygon_instance_ps = (spPolygonInstancePS)polygon_instance;

    spDebug(50, "spDrawPolygonPS", "dx = %d, dy = %d\n", polygon_instance_ps->dx, polygon_instance_ps->dy);
    
    updateGraphicParams(pinstance, graphics);
    return spPSDrawPolygon(pinstance->psfile, polygon_instance_ps->polygon_id,
			   polygon_instance_ps->dx, polygon_instance_ps->dy);
}

static char *sp_drawing_plugin_ps_type_list[] = {
    "PS",
    "EPS",
    NULL,
};

static char *sp_drawing_plugin_ps_desc_list[] = {
    "Postscript",
    "Encapsuled Postscript",
    NULL,
};

static char *sp_drawing_plugin_ps_filter_list[] = {
    "*.ps",
    "*.eps",
    NULL,
};

static spDrawingPluginRec sp_drawing_plugin_ps = {
    NULL,
    NULL,

    SP_PLUGIN_DRAWING,
    SP_DRAWING_PLUGIN_PS,
    1,
    SP_PLUGIN_PRIORITY_MIDDLE,
    SP_DRAWING_PLUGIN_CAPS_MULTIPLE_PAGES,
    spInitPS,
    spFreePS,
    "PostScript File",
    "PostScript Drawing Plugin Version 0.1\nCopyright(C) Hideki Banno",
    
    spInitInstancePS,
    spFreeInstancePS,
    NULL,
    NULL,

    SP_DRAWING_DEVICE_FILE,

    NULL,
    NULL,
    NULL,
    NULL,

    NULL,
    spCreateComponentPS,
    spDestroyComponentPS,
    NULL/*spSetParamsPS*/,
    NULL/*spGetParamsPS*/,

    NULL,

    spSetPageInfoPS,
    spGetPageInfoPS,
    NULL,
    spGetPageRangePS,
    NULL,
    
    sp_drawing_plugin_ps_type_list,
    sp_drawing_plugin_ps_desc_list,
    sp_drawing_plugin_ps_filter_list,
    spSetFileTypePS,
    NULL,
    
    spOpenTargetFilePS,
    NULL,

    spCloseTargetFilePS,

    spIsDrawablePS,

    spGetDrawingCapsPS,
    spGetGraphicsModeCapsPS,
    NULL,
    spGetClientSizePS,
    spGetPhysicalSizePS,
    spSetScalingFactorPS,

    spStartDrawingPS,
    spEndDrawingPS,
    spForwardPagePS,

    NULL,
    NULL,

    spFillRectanglePS,
    spDrawRectanglePS,
    spFillArcPS,
    spDrawArcPS,
    spDrawLinePS,
    spDrawPointPS,
    spDrawStringPS,
    NULL,

    NULL,
    spInitPolygonInstancePS,
    spFreePolygonInstancePS,
    
    spPolygonEventPS,
    spFillPolygonPS,
    spDrawPolygonPS,
};

spBool spAddPSPlugin(void)
{
    return spAddPluginRecList((spPluginRec *)&sp_drawing_plugin_ps, SP_DRAWING_PLUGIN_PS);
}

spBool spRemovePSPlugin(void)
{
    return spRemovePluginRecList((spPluginRec *)&sp_drawing_plugin_ps);
}

#if (defined(GTK) && !defined(GTK3)) || defined(USE_MOTIF) || defined(ANDROID)
static spBool spOpenTargetDialogPSPrinter(void *instance, spArg *args, int num_arg)
{
    char *printer_name;
    int start_page, end_page;
    int num_copies;
    spBool collate;
    spBool double_sided, tumble;
    static long id = 0;
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;
    
    if ((printer_name = xspOpenPrinterDialogDefault(&start_page, &end_page, &num_copies, &collate,
						    &double_sided, &tumble, args, num_arg)) == NULL) {
	return SP_FALSE;
    }
    spDebug(50, "spOpenTargetDialogPSPrinter", "printer_name = %s\n", printer_name);
    spDebug(50, "spOpenTargetDialogPSPrinter", "start_page = %d, end_page = %d\n", start_page, end_page);
    spDebug(50, "spOpenTargetDialogPSPrinter", "num_copies = %d, collate = %d\n", num_copies, collate);
    spDebug(50, "spOpenTargetDialogPSPrinter", "double_sided = %d, tumble = %d\n", double_sided, tumble);

    if (double_sided == SP_TRUE) {
	pinstance->options |= SP_PS_DOUBLE_SIDED;
	if (tumble == SP_TRUE) {
	    pinstance->options |= SP_PS_TUMBLE;
	}
    }

    sprintf(pinstance->temp_file, "%s%csppsplugin%ld-%ld.ps", spGetApplicationTempDir(), SP_DIR_SEPARATOR,
	    spGetProcessId(), id++);

    if ((pinstance->psfile = spPSOpenFile(pinstance->temp_file, pinstance->width_mm,
					  pinstance->height_mm, pinstance->options)) != NULL) {
	if (pinstance->printer_name != NULL) xfree(pinstance->printer_name);
	pinstance->printer_name = printer_name;

	spPSSetNumCopies(pinstance->psfile, num_copies, collate);
    
	pinstance->start_page = start_page;
	pinstance->end_page = end_page;
    
	return SP_TRUE;
    } else {
	xfree(printer_name);
	
	return SP_FALSE;
    }
}

static spBool spCloseTargetPSPrinter(void *instance)
{
    static char printer[SP_MAX_LINE];
    static char command[SP_MAX_MESSAGE];
    spBool flag = SP_FALSE;
    spPluginInstancePS pinstance = (spPluginInstancePS)instance;

    if (pinstance->psfile != NULL) {
	spPSCloseFile(pinstance->psfile);
	
	printer[0] = '\0';
	if (!strnone(pinstance->printer_name)) {
	    sprintf(printer, "-P%s", pinstance->printer_name);
	}

	sprintf(command, "lpr %s \"%s\"", printer, pinstance->temp_file);
	spDebug(50, "spCloseTargetPSPrinter", "command = %s\n", command);

#if 1
	if (system(command) == 0) {
	    flag = SP_TRUE;
	}
#endif
    }
    pinstance->psfile = NULL;

    spRemoveFile(pinstance->temp_file);
    pinstance->temp_file[0] = '\0';
    
    pinstance->start_page = 0;
    pinstance->end_page = 0;
    
    pinstance->options &= ~SP_PS_DOUBLE_SIDED;
    pinstance->options &= ~SP_PS_TUMBLE;
    
    return flag;
}

static spDrawingPluginRec sp_drawing_plugin_printer = {
    NULL,
    NULL,

    SP_PLUGIN_DRAWING,
    SP_DRAWING_PLUGIN_PRINTER,
    1,
    SP_PLUGIN_PRIORITY_MIDDLE,
    0L,	/* caps */
    spInitPS,
    spFreePS,
    "PostScript Printer",
    "PostScript Printer Plugin for UNIX  Version 0.1\nCopyright(C) Hideki Banno\nE-mail: banno@ccmfs.meijo-u.ac.jp",
    
    spInitInstancePS,
    spFreeInstancePS,
    NULL,
    NULL,

    SP_DRAWING_DEVICE_PRINTER,

    NULL,
    NULL,
    NULL,
    NULL,

    NULL,
    spCreateComponentPS,
    spDestroyComponentPS,
    NULL/*spSetParamsPS*/,
    NULL/*spGetParamsPS*/,

    NULL,

    spSetPageInfoPS,
    NULL,
    NULL,
    NULL,
    NULL,
    
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    
    NULL,
    spOpenTargetDialogPSPrinter,

    spCloseTargetPSPrinter,

    spIsDrawablePS,

    spGetDrawingCapsPS,
    spGetGraphicsModeCapsPS,
    NULL,
    spGetClientSizePS,
    spGetPhysicalSizePS,
    spSetScalingFactorPS,

    spStartDrawingPS,
    spEndDrawingPS,
    spForwardPagePS,

    NULL,
    NULL,

    spFillRectanglePS,
    spDrawRectanglePS,
    spFillArcPS,
    spDrawArcPS,
    spDrawLinePS,
    spDrawPointPS,
    spDrawStringPS,
    NULL,

    NULL,
    spInitPolygonInstancePS,
    spFreePolygonInstancePS,
    
    spPolygonEventPS,
    spFillPolygonPS,
    spDrawPolygonPS,
};

spBool spAddPrinterPlugin(void)
{
    return spAddPluginRecList((spPluginRec *)&sp_drawing_plugin_printer, SP_DRAWING_PLUGIN_PRINTER);
}

spBool spRemovePrinterPlugin(void)
{
    return spRemovePluginRecList((spPluginRec *)&sp_drawing_plugin_printer);
}
#endif
