/*
 *	spFrame_Win.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spButton.h>

#include <sp/spTopLevelP.h>
#include <sp/spFrameP.h>

extern spTopLevel sp_toplevel;

static spBool sp_main_frame_created = SP_FALSE;

void spFramePartInitArch(spComponent component)
{
    SpFrameArch(component).task_tray_added = SP_FALSE;
    memset(&SpFrameArch(component).nicon, 0, sizeof(NOTIFYICONDATA));
    SpFrameArch(component).hfocus_text = NULL;
    SpFrameArch(component).drop_call_func = NULL;
    SpFrameArch(component).drop_call_data = NULL;
    SpFrameArch(component).prop_set = SP_FALSE;
    
    return;
}

void spFramePartFreeArch(spComponent component)
{
    return;
}

static void getDesktopCenter(int *x, int *y)
{
    HWND hdesk;
    RECT rect;
    
    hdesk = GetDesktopWindow();
    GetClientRect(hdesk, &rect);

    if (x != NULL) {
	*x = (rect.right - rect.left) / 2;
    }
    if (x != NULL) {
	*y = (rect.bottom - rect.top) / 2;
    }

    return;
}

static void getDesktopSize(int *width, int *height)
{
    HWND hdesk;
    RECT rect;
    
    hdesk = GetDesktopWindow();
    GetClientRect(hdesk, &rect);

    if (width != NULL) {
	*width = (rect.right - rect.left);
    }
    if (height != NULL) {
	*height = (rect.bottom - rect.top);
    }

    return;
}

void spFrameCreateArch(spComponent component)
{
    int width;
    int height;
    char *title = NULL;
    char *icon_name = NULL;
    DWORD dwexstyle = WS_EX_CONTROLPARENT;
    DWORD dwstyle = WS_CLIPCHILDREN;
    HMENU hsysmenu = NULL;
    HWND hparent = HWND_DESKTOP;

    if (SpComponentPart(component).width > 0) {
	width = spScaleXWin(sp_toplevel, SpComponentPart(component).width);
    } else {
	width = spScaleXWin(sp_toplevel, SP_DEFAULT_WINDOW_WIDTH);
    }
    if (SpComponentPart(component).height > 0) {
	height = spScaleYWin(sp_toplevel, SpComponentPart(component).height);
    } else {
	height = spScaleYWin(sp_toplevel, SP_DEFAULT_WINDOW_HEIGHT);
    }
    
    if (!strnone(SpFramePart(component).icon_name)) {
	icon_name = SpFramePart(component).icon_name;
    }
    
    if (!strnone(SpComponentPart(component).title)) {
	title = SpComponentPart(component).title;
    }

    if (title == NULL) {
	if (icon_name != NULL) {
	    title = icon_name;
	} else {
	    title = (!strnone(SpGetName(component)) ? SpGetName(component) : "");
	}
    }
    
    if (SpFramePart(component).system_window != NULL) {
	hparent = (HWND)SpFramePart(component).system_window;
	dwstyle |= WS_CHILD | WS_CLIPSIBLINGS;
	spDebug(50, "spFrameCreateArch", "using system_window: dwstyle = %lx\n", dwstyle);
	SpTopLevelArch(sp_toplevel).getting_message = SP_TRUE;
    } else {
	if (spIsSubClass(component, SpMainFrame) == SP_TRUE
	    && sp_main_frame_created == SP_FALSE) {
	    sp_main_frame_created = SP_TRUE;
	}
    
	if (SpFramePart(component).window_type == SP_DIALOG_WINDOW) {
	    dwstyle |= WS_CAPTION | WS_SYSMENU;
	} else if (SpFramePart(component).window_type == SP_TRANSIENT_WINDOW
		   || SpFramePart(component).window_type == SP_FULLSCREEN_WINDOW) {
	    dwstyle |= WS_POPUP | WS_CLIPCHILDREN;
	    dwexstyle |= WS_EX_TOOLWINDOW;
	} else {
	    dwstyle |= WS_CAPTION | WS_SYSMENU;
	}
    
	if (SpFramePart(component).window_type != SP_TRANSIENT_WINDOW
	    && SpFramePart(component).window_type != SP_FULLSCREEN_WINDOW) {
	    if (SpFramePart(component).resize_flag == SP_TRUE) {
		dwstyle |= WS_MAXIMIZEBOX | WS_SIZEBOX | WS_SYSMENU;
	    } else {
		dwstyle &= (~WS_MAXIMIZEBOX);
		dwstyle &= (~WS_SIZEBOX);
	    }
	    if (SpFramePart(component).iconfy_flag == SP_TRUE) {
		dwstyle |= WS_MINIMIZEBOX | WS_SYSMENU;
	    } else {
		dwstyle &= (~WS_MINIMIZEBOX);
	    }
	    spDebug(50, "spFrameCreateArch", "resize_flag = %d, iconfy_flag = %d, help_button_visible = %d\n",
		    SpFramePart(component).resize_flag, SpFramePart(component).iconfy_flag,
		    SpFramePart(component).help_button_visible);
	    if (SpFramePart(component).resize_flag == SP_FALSE
		&& SpFramePart(component).iconfy_flag == SP_FALSE
		&& SpFramePart(component).help_button_visible == SP_TRUE) {
		dwexstyle |= WS_EX_CONTEXTHELP;
	    }
	    if (SpFramePart(component).topmost_flag == SP_TRUE) {
		dwexstyle |= WS_EX_TOPMOST;
	    }
	}

	if (SpFramePart(component).window_type == SP_FULLSCREEN_WINDOW) {
#if 0
	    RECT rect;
	    SystemParametersInfo(SPI_GETWORKAREA, 0, &rect, 0);
	    width = rect.right - rect.left;
	    height = rect.bottom - rect.top;
#else
	    /* 2012/8  since current version unmaps task bar */
	    width = GetSystemMetrics(SM_CXSCREEN);
	    height = GetSystemMetrics(SM_CYSCREEN);
#endif
	    SpComponentPart(component).current_width = width;
	    SpComponentPart(component).current_height = height;
	    SpComponentPart(component).x = 0;
	    SpComponentPart(component).y = 0;
	} else if (SpComponentPart(component).x == 0 && SpComponentPart(component).y == 0) {
	    int dwidth, dheight;
	    static int x = -1;
	    static int y = -1;
	
	    getDesktopSize(&dwidth, &dheight);
	
	    if (x < 0 || x > dwidth / 3 || y > dheight / 3) {
		x = spScaleXWin(sp_toplevel, SP_INIT_WINDOW_X_POS);
		y = spScaleYWin(sp_toplevel, SP_INIT_WINDOW_Y_POS);
	    } else {
		x += spScaleXWin(sp_toplevel, SP_WINDOW_X_INCR);
		y += spScaleYWin(sp_toplevel, SP_WINDOW_Y_INCR);
	    }
	    SpComponentPart(component).x = x;
	    SpComponentPart(component).y = y;
	}
    }
    
    SpPrimitiveArch(component).hwnd =
	CreateWindowEx(dwexstyle,
		       SP_WINDOW_CLASS_NAME,
		       title,
		       dwstyle,
		       SpComponentPart(component).x,
		       SpComponentPart(component).y,
		       width, height,
		       hparent,
		       NULL,
		       SpTopLevelArch(sp_toplevel).hThisInst,
		       NULL);
    if ((SpFramePart(component).window_type != SP_TRANSIENT_WINDOW
	 && SpFramePart(component).window_type != SP_FULLSCREEN_WINDOW)
	&& SpFramePart(component).system_window == NULL) {
	hsysmenu = GetSystemMenu(SpPrimitiveArch(component).hwnd, FALSE);
	if (SpFramePart(component).close_style == SP_NO_CLOSE) {
	    EnableMenuItem(hsysmenu, SC_CLOSE,
			   MF_BYCOMMAND | MF_GRAYED);
	}
	if (SpFramePart(component).resize_flag == SP_FALSE) {
	    EnableMenuItem(hsysmenu, SC_SIZE,
			   MF_BYCOMMAND | MF_GRAYED);
	    EnableMenuItem(hsysmenu, SC_MAXIMIZE,
			   MF_BYCOMMAND | MF_GRAYED);
	}
    }
    if (SpFramePart(component).window_type == SP_FULLSCREEN_WINDOW) {
	SetWindowPos(SpPrimitiveArch(component).hwnd, /*HWND_TOPMOST*/HWND_TOP,
		     0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
    }
	
    if (!strnone(SpTopLevelPart(sp_toplevel).unique_id)) {
	SetProp(SpPrimitiveArch(component).hwnd, SpTopLevelPart(sp_toplevel).unique_id,
		(HANDLE)GetCurrentProcessId());
	SpFrameArch(component).prop_set = SP_TRUE;
    }
    
    return;
}

void spFrameDestroyArch(spComponent component)
{
    if (SpFrameArch(component).prop_set == SP_TRUE) {
	SpFrameArch(component).prop_set = SP_FALSE;
	RemoveProp(SpPrimitiveArch(component).hwnd, SpTopLevelPart(sp_toplevel).unique_id);
    }
    return;
}
    
void spFrameSetParamsArch(spComponent component)
{
    char *title = NULL;
    char *icon_name = NULL;
    HMENU hsysmenu = NULL;

    if (SpFramePart(SpOldObject(component)).icon_name != SpFramePart(component).icon_name) {    
	if (!strnone(SpFramePart(component).icon_name)) {
	    icon_name = SpFramePart(component).icon_name;
	}
    }
    
    if (SpFramePart(SpOldObject(component)).close_style != SpFramePart(component).close_style
	&& SpFramePart(component).system_window == NULL) {
	hsysmenu = GetSystemMenu(SpPrimitiveArch(component).hwnd, FALSE);
	if (SpFramePart(component).close_style == SP_NO_CLOSE) {
	    EnableMenuItem(hsysmenu, SC_CLOSE,
			   MF_BYCOMMAND | MF_GRAYED);
	} else {
	    EnableMenuItem(hsysmenu, SC_CLOSE,
			   MF_BYCOMMAND | MF_ENABLED);
	}
    }
    
    if (SpFramePart(SpOldObject(component)).topmost_flag != SpFramePart(component).topmost_flag
	&& SpFramePart(component).system_window == NULL) {
	if (SpFramePart(component).topmost_flag == SP_TRUE) {
	    SetWindowPos(SpPrimitiveArch(component).hwnd, HWND_TOPMOST,
			 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
	} else {
	    SetWindowPos(SpPrimitiveArch(component).hwnd, HWND_NOTOPMOST,
			 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
	}
    }
	
    if (SpComponentPart(SpOldObject(component)).title != SpComponentPart(component).title) {    
	if (!strnone(SpComponentPart(component).title)) {
	    title = SpComponentPart(component).title;
	}
    }

    if (title != NULL) {
	SendMessage(SpPrimitiveArch(component).hwnd, WM_SETTEXT,
		    (WPARAM)0, (LPARAM)(LPCTSTR)title);
    } else if (icon_name != NULL) {
	SendMessage(SpPrimitiveArch(component).hwnd, WM_SETTEXT,
		    (WPARAM)0, (LPARAM)(LPCTSTR)icon_name);
    }

    return;
}

void *spGetArchFrame(spComponent component)
{
    return (void *)SpPrimitiveArch(component).hwnd;
}

spBool spIsTaskTraySupported(void)
{
    return SP_TRUE;
}

spBool spShowTaskTrayArch(spComponent window, spBool flag)
{
    char *icon_name;
    const char *message;
    HICON hIcon;

    spDebug(10, "spShowTaskTrayArch", "in\n");

    SpFrameArch(window).nicon.cbSize = sizeof(NOTIFYICONDATA);
    SpFrameArch(window).nicon.uID = 1;
    
    if (flag == SP_TRUE) {
	if (SpPrimitiveArch(window).realize_flag == SP_FALSE) {
	    if (SpPrimitiveArch(window).hmenu != NULL) {
		SetMenu(SpPrimitiveArch(window).hwnd, SpPrimitiveArch(window).hmenu);
	    }
	    ShowWindow(SpPrimitiveArch(window).hwnd, SW_HIDE);
	    if (SpFramePart(window).window_type == SP_TRANSIENT_WINDOW
		|| SpFramePart(window).window_type == SP_FULLSCREEN_WINDOW) {
		spAdjustComponentSize(window);
	    }
	    SpPrimitiveArch(window).realize_flag = SP_TRUE;
	    UpdateWindow(SpPrimitiveArch(window).hwnd);
	}
	
	if (!strnone(SpFramePart(window).task_tray_icon)) {
	    icon_name = SpFramePart(window).task_tray_icon;
	} else if (!strnone(SpTopLevelPart(sp_toplevel).icon_name)) {
	    icon_name = SpTopLevelPart(sp_toplevel).icon_name;
	} else {
	    icon_name = NULL;
	}

	if (icon_name != NULL) {
	    spDebug(10, "spShowTaskTrayArch", "icon_name = %s\n", icon_name);
	    if ((hIcon = LoadIcon(SpTopLevelArch(sp_toplevel).hThisInst,
				  icon_name)) == NULL) {
		return SP_FALSE;
	    }
	} else {
	    if ((hIcon = LoadIcon(NULL, IDI_WINLOGO)) == NULL) {
		return SP_FALSE;
	    }
	}
	
	SpFrameArch(window).nicon.hWnd = SpPrimitiveArch(window).hwnd;
	SpFrameArch(window).nicon.uFlags = NIF_ICON | NIF_MESSAGE;
	SpFrameArch(window).nicon.hIcon = hIcon;
	SpFrameArch(window).nicon.uCallbackMessage = WM_TASKTRAY;

	if (!strnone(SpFramePart(window).task_tray_tip)) {
	    message = SpFramePart(window).task_tray_tip;
	} else if (!strnone(SpComponentPart(window).description)) {
	    message = SpComponentPart(window).description;
	} else {
	    message = spGetTitle(window);
	}
	if (!strnone(message)) {
	    spDebug(10, "spShowTaskTrayArch", "tip = %s\n", message);
	    SpFrameArch(window).nicon.uFlags |= NIF_TIP;
	    strcpy(SpFrameArch(window).nicon.szTip, message);
	}
	
	if (SpFrameArch(window).task_tray_added == SP_FALSE) {
	    Shell_NotifyIcon(NIM_ADD, &SpFrameArch(window).nicon);
	    SpFrameArch(window).task_tray_added = SP_TRUE;
	} else {
	    Shell_NotifyIcon(NIM_MODIFY, &SpFrameArch(window).nicon);
	}
    } else {
	if (SpFrameArch(window).task_tray_added == SP_TRUE) {
	    Shell_NotifyIcon(NIM_DELETE, &SpFrameArch(window).nicon);
	    SpFrameArch(window).task_tray_added = SP_FALSE;
	}
    }

    return SP_TRUE;
}

static spComponent sp_grabbed_window = NULL;

void spRemoveGrabWin(spComponent window)
{
    spComponent first_window, next_window;
    
    if (sp_grabbed_window != NULL
	&& (window == NULL || sp_grabbed_window == window)) {
	window = sp_grabbed_window;

	next_window = spGetNextWindow(window, SP_TRUE);
	first_window = next_window;
	while (next_window != window && spIsCreated(next_window) == SP_TRUE) {
	    if (spIsPrimitive(next_window) == SP_TRUE
		&& SpPrimitiveArch(next_window).hwnd != NULL) {
		EnableWindow(SpPrimitiveArch(next_window).hwnd,
			     SpComponentPart(next_window).sensitive_flag);
	    }
	    next_window = spGetNextWindow(next_window, SP_TRUE);

	    if (next_window == first_window) {
		break;
	    }
	}
	
	sp_grabbed_window = NULL;
    }
    
    return;
}

spComponent spAddGrabWin(spComponent window)
{
    spComponent prev = NULL;
    spComponent next_window;

    if (sp_grabbed_window != NULL) {
	prev = sp_grabbed_window;
	spRemoveGrabWin(NULL);
    }
    
    if (spIsCreated(window) == SP_TRUE) {
	next_window = spGetNextWindow(window, SP_TRUE);
	while (next_window != window && spIsCreated(next_window) == SP_TRUE) {
	    if (spIsPrimitive(next_window) == SP_TRUE
		&& SpPrimitiveArch(next_window).hwnd != NULL) {
		EnableWindow(SpPrimitiveArch(next_window).hwnd, FALSE);
	    }
	    next_window = spGetNextWindow(next_window, SP_TRUE);
	}
	
	sp_grabbed_window = window;
    }
    
    return prev;
}

spComponent spGetCurrentGrabWin(void)
{
    return sp_grabbed_window;
}

static spBool setFramePosition(spComponent window, int *x, int *y)
{
    RECT rect;
    int width, height;

    getDesktopSize(&width, &height);
    
    while (*x >= width) {
	*x -= width;
    }
    while (*x < 0) {
	*x += width;
    }
    
    while (*y >= height) {
	*y -= height;
    }
    while (*y < 0) {
	*y += height;
    }
    spDebug(10, "setFramePosition", "x = %d, y = %d\n", *x, *y);
    
    GetWindowRect(SpPrimitiveArch(window).hwnd, &rect);
    MoveWindow(SpPrimitiveArch(window).hwnd,
	       *x, *y, rect.right - rect.left, rect.bottom - rect.top, TRUE);

    return SP_TRUE;
}

void spPopupFrameArch(spComponent window, spBool move_cursor)
{
    int nWinMode = SW_SHOW;
    spDebug(30, "spPopupFrameArch", "in\n");
    
    if (SpFramePart(window).window_type == SP_FULLSCREEN_WINDOW) {
	spUnmapTaskBarWindow(sp_toplevel);
    }
    
    if (SpPrimitiveArch(window).realize_flag == SP_FALSE) {
	if (SpPrimitiveArch(window).hmenu != NULL) {
	    SetMenu(SpPrimitiveArch(window).hwnd, SpPrimitiveArch(window).hmenu);
	}
	if (SpFramePart(window).system_window == NULL) {
	    nWinMode = SpTopLevelArch(sp_toplevel).nWinMode;
	}

	ShowWindow(SpPrimitiveArch(window).hwnd, nWinMode);
	spDebug(50, "spPopupFrameArch", "show window done\n");
	
	if ((SpFramePart(window).window_type == SP_TRANSIENT_WINDOW
	     || SpFramePart(window).window_type == SP_FULLSCREEN_WINDOW)
	    || SpFramePart(window).system_window != NULL) {
	    spAdjustComponentSize(window);
	}
	if (spIsButton(SpFramePart(window).default_button) == SP_TRUE) {
	    SendMessage(SpPrimitiveArch(SpFramePart(window).default_button).hwnd,
			BM_SETSTYLE, (WPARAM)BS_DEFPUSHBUTTON, TRUE);
	}
	
	SpPrimitiveArch(window).realize_flag = SP_TRUE;
    } else {
	if (SpFramePart(window).system_window == NULL) {
	    ShowWindow(SpPrimitiveArch(window).hwnd, SW_RESTORE);
	    SetForegroundWindow(SpPrimitiveArch(window).hwnd);
	} else {
	    ShowWindow(SpFramePart(window).system_window, SW_RESTORE);
	    SetForegroundWindow(SpFramePart(window).system_window);
	}
    }
    UpdateWindow(SpPrimitiveArch(window).hwnd);
    
    if (SpFramePart(window).task_tray_visible == SP_TRUE
	&& SpFrameArch(window).task_tray_added == SP_FALSE) {
	spShowTaskTrayArch(window, SpFramePart(window).task_tray_visible);
    }

    if (SpFramePart(window).popup_style == SP_MODAL_POPUP || SpFramePart(window).popup_style == SP_SYSTEM_MODAL_POPUP) {
	HWND hwnd;
	spComponent prev_grabbed;
	
	prev_grabbed = spAddGrabWin(window);
	
	hwnd = SpPrimitiveArch(window).hwnd;
	
	spDebug(30, "spPopupFrameArch", "loop in\n");
	while (1) {
	    spWaitNextEvent(SpFramePart(window).toplevel);
	    
	    if (hwnd == NULL || IsWindowVisible(hwnd) == FALSE) {
		break;
	    }
	}
	spDebug(30, "spPopupFrameArch", "loop end\n");

	spAddGrabWin(prev_grabbed);
    }
	
    spDebug(30, "spPopupFrameArch", "done\n");
    
    return;
}

void spPopdownFrameArch(spComponent window)
{
    spRemoveGrabWin(NULL);
    
    if (SpFramePart(window).window_type == SP_FULLSCREEN_WINDOW) {
	spMapTaskBarWindow(sp_toplevel);
    }
    
    if (SpFramePart(window).system_window != NULL) {
	ShowWindow(SpFramePart(window).system_window, SW_HIDE);
    } else if (SpPrimitiveArch(window).hwnd != NULL) {
	ShowWindow(SpPrimitiveArch(window).hwnd, SW_HIDE);
    }
    
    spDebug(30, "spPopdownFrameArch", "done\n");

    return;
}

void spRaiseFrameArch(spComponent window)
{
    if (SpFramePart(window).window_type == SP_FULLSCREEN_WINDOW) {
	spUnmapTaskBarWindow(sp_toplevel);
	SetWindowPos(SpPrimitiveArch(window).hwnd, /*HWND_TOPMOST*/HWND_TOP,
		     0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
    } else {
	spMapTaskBarWindow(sp_toplevel);
	SetWindowPos(SpPrimitiveArch(window).hwnd, HWND_TOP,
		     0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
    }
    SetForegroundWindow(SpPrimitiveArch(window).hwnd);
    return;
}

void spLowerFrameArch(spComponent window)
{
    if (SpFramePart(window).window_type == SP_FULLSCREEN_WINDOW) {
	spMapTaskBarWindow(sp_toplevel);
    }
    
    SetWindowPos(SpPrimitiveArch(window).hwnd, HWND_BOTTOM,
		 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
    
    return;
}

spBool spSetFramePositionArch(spComponent window, int x, int y)
{
    return setFramePosition(window, &x, &y);
}

spBool spGetFramePositionArch(spComponent window, int *x, int *y)
{
    RECT rect;
    
    GetWindowRect(SpPrimitiveArch(window).hwnd, &rect);
    
    *x = rect.left;
    *y = rect.top;
    
    return SP_TRUE;
}

spBool spGetFrameModifierKeyMaskArch(spComponent window, spModifierMask *mask)
{
    *mask = 0L;
    
    if (GetAsyncKeyState(VK_SHIFT) & 0x8000) {
	*mask |= SP_SHIFT_MASK;
    }
    if (GetAsyncKeyState(VK_CONTROL) & 0x8000) {
	*mask |= SP_CONTROL_MASK;
    }
    if (GetAsyncKeyState(VK_MENU) & 0x8000) {
	*mask |= SP_ALT_MASK;
    }
    if (LOBYTE(GetKeyState(VK_CAPITAL))) {
	*mask |= SP_LOCK_MASK;
    }
    
    return SP_TRUE;
}

spTimerId spSetTimerArch(spTimerList timer)
{
    UINT_PTR timer_id = 1;

    if (timer->prev != NULL) {
	timer_id = (UINT_PTR)(timer->prev->timer_id) + 1;
    }
    
    timer_id = SetTimer(SpPrimitiveArch(timer->frame).hwnd,
			timer_id, (UINT)timer->interval_ms, NULL);
	     
    return (spTimerId)timer_id;
}

spBool spRemoveTimerWin(spTimerList timer)
{
    if (KillTimer(SpPrimitiveArch(timer->frame).hwnd, (UINT_PTR)timer->timer_id)) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}
    
spBool spKillTimerArch(spTimerList timer)
{
    /* killed in message function */
    timer->kill_flag = SP_TRUE;
    return SP_TRUE;
}

spBool spAddDropCallbackArch(spComponent component, spDropCallbackFunc call_func, void *call_data)
{
    if (call_func == NULL) {
	DragAcceptFiles(SpPrimitiveArch(component).hwnd, FALSE);
	SpFrameArch(component).drop_call_func = NULL;
	SpFrameArch(component).drop_call_data = NULL;
    } else {
	if (SpFrameArch(component).drop_call_func == NULL) {
	    DragAcceptFiles(SpPrimitiveArch(component).hwnd, TRUE);
	}
	SpFrameArch(component).drop_call_func = call_func;
	SpFrameArch(component).drop_call_data = call_data;
    }

    return SP_TRUE;
}

HINSTANCE spGetCurrentInstanceWin(spComponent component)
{
    spTopLevel toplevel;

    if ((toplevel = spGetTopLevel(component)) != NULL) {
	return SpTopLevelArch(toplevel).hThisInst;
    } else {
	return NULL;
    }
}
