/*
 *	spImage_Xm.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xlocale.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/xpm.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spDrawP.h>
#include <sp/spImageP.h>
#include <sp/spGraphicsP.h>

extern spTopLevel sp_toplevel;

spBool spSavePixmapToFileXm(Pixmap pixmap, const char *filename)
{
    if (XpmWriteFileFromPixmap(SpTopLevelArch(sp_toplevel).display,
			       filename,
			       pixmap, 
			       None, 
			       NULL) == XpmSuccess) {
	spDebug(30, "spSavePixmapToFileXm", "pixmap write done: %s\n", filename);
	return SP_TRUE;
    } else {
	spDebug(10, "spSavePixmapToFileXm", "pixmap write failed\n");
	return SP_FALSE;
    }
}

Pixmap spCreatePixmapFromDataOrFileXm(Widget widget, char **bitmap_data, const char *basename,
				      int *width, int *height)
{
    Pixmap pixmap = None;
    Display *display;
    Screen *screen;
    XWindowAttributes root_attr;
    XpmAttributes xpm_attr;
    XpmColorSymbol xpm_symbols[1];
    Pixel color;
    /*Pixmap mask_bitmap;*/
    char filename[SP_MAX_PATHNAME];

    if (widget != NULL) {
	display = XtDisplay(widget);
	screen = XtScreen(widget);
	XtVaGetValues(widget,
		      XmNbackground, &color,
		      NULL);
	spDebug(30, "spCreatePixmapFromDataOrFileXm", "non-NULL widget case: color = %lx\n", color);
    } else {
	display = SpTopLevelArch(sp_toplevel).display;
	screen = SpTopLevelArch(sp_toplevel).screen;
	color = spGetColorPixelXm("#c0c0c0");
	spDebug(30, "spCreatePixmapFromDataOrFileXm", "NULL widget case: color = %lx\n", color);
    }
    spDebug(30, "spCreatePixmapFromDataOrFileXm", "display = %ld, screen = %ld\n", display, screen);

    XGetWindowAttributes(display, RootWindowOfScreen(screen),
			 &root_attr);
    xpm_attr.colormap = root_attr.colormap;
    spDebug(30, "spCreatePixmapFromDataOrFileXm", "colormap = %ld\n", root_attr.colormap);
    xpm_attr.closeness = 40000;
    xpm_attr.depth = DefaultDepthOfScreen(screen);
    xpm_attr.valuemask = XpmSize | XpmReturnPixels | XpmColormap | XpmCloseness | XpmDepth;

    xpm_symbols[0].name = NULL;
    xpm_symbols[0].value = "none";
    xpm_symbols[0].pixel = color;
    xpm_attr.colorsymbols = xpm_symbols;
    xpm_attr.numsymbols = 1;
    xpm_attr.valuemask |= XpmColorSymbols;

    if (bitmap_data != NULL) {
	if (XpmCreatePixmapFromData(display,
				    RootWindowOfScreen(screen),
				    bitmap_data,
				    &pixmap, 
				    /*&mask_bitmap*/NULL, 
				    &xpm_attr) == XpmSuccess) {
	    if (width != NULL) *width = xpm_attr.width;
	    if (height != NULL) *height = xpm_attr.height;
	    XpmFreeAttributes(&xpm_attr);
	    spDebug(30, "spCreatePixmapFromDataOrFileXm", "pixmap create done: pixmap = %ld\n", pixmap);
	}
    } else {
	if (!strnone(basename)) {
	    if (spEqSuffix(basename, ".xpm")) {
		strcpy(filename, basename);
	    } else {
		sprintf(filename, "%s.xpm", basename);
	    }
	    spDebug(30, "spCreatePixmapFromDataOrFileXm", "filename = %s\n", filename);
	    
	    if (XpmReadFileToPixmap(display,
				    RootWindowOfScreen(screen),
				    filename,
				    &pixmap, 
				    /*&mask_bitmap*/NULL, 
				    &xpm_attr) == XpmSuccess) {
		if (width != NULL) *width = xpm_attr.width;
		if (height != NULL) *height = xpm_attr.height;
		XpmFreeAttributes(&xpm_attr);
		spDebug(30, "spCreatePixmapFromDataOrFileXm", "pixmap load done: %s, pixmap = %ld\n",
			filename, pixmap);
	    }
	}
    }
    
    return pixmap;
}

void spImageCreateArch(spComponent component)
{
    return;
}

spBool spImageLoadArch(spComponent component, const char *filename)
{
    SpPrimitiveArch(component).pixmap = spCreatePixmapFromDataOrFileXm(SpParentPrimitiveArch(component).widget,
								       NULL, filename,
								       &SpComponentPart(component).width, &SpComponentPart(component).height);
    if (SpPrimitiveArch(component).pixmap != None) {
	return SP_TRUE;
    }
    
    return SP_FALSE;
}

void spDrawImageArch(spComponent component)
{
    spGetDisplayInfo(component);
    
    XFillRectangle(SpPrimitiveArch(component).display, SpPrimitiveArch(component).pixmap,
		   SpTopLevelArch(sp_toplevel).bg_gc,
		   0, 0, SpComponentPart(component).client_width,
		   SpComponentPart(component).client_height);
    
    return;
}

void spRedrawImageArch(spComponent component)
{
    if (spIsSubClass(component, SpImage) == SP_TRUE
	&& !strnone(SpComponentPart(component).description)) return;
    
    spGetDisplayInfo(component);
    
    if (SpPrimitiveArch(component).pixmap != None) {
	XFreePixmap(SpPrimitiveArch(component).display, SpPrimitiveArch(component).pixmap);
	SpPrimitiveArch(component).pixmap = None;
    }
    
    /* create pixmap */
    if ((SpPrimitiveArch(component).pixmap = 
	 XCreatePixmap(SpPrimitiveArch(component).display, 
		       RootWindowOfScreen(SpPrimitiveArch(component).screen),
		       SpComponentPart(component).client_width, SpComponentPart(component).client_height,
		       DefaultDepthOfScreen(SpPrimitiveArch(component).screen))) == None) {
	spError(1, "can't create pixmap\n");
    }
    
    return;
}

