/*
 *	spList.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spContainer.h>
#include <sp/spComboBox.h>

#include <sp/spListP.h>

static spParamTable sp_list_param_tables[] = {
    {SppListStrings, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spList, list.strings), NULL},
    {SppSelectedListIndex, SP_CREATE_ACCESS/* | SP_SET_ACCESS | SP_GET_ACCESS*/,
	 spOffset(spList, primitive.index), "-1"},
};

spListClassRec SpListClassRec = {
    /* spObjectClassPart */
    {
	SpList,
	(spObjectClass)&SpPrimitiveClassRec,
	sizeof(spListRec),
	spArraySize(sp_list_param_tables),
	sp_list_param_tables,
	spListPartInit,
	spListPartFree,
	SP_FALSE,
	NULL,
	NULL,
	spListCreate,
	spListDestroy,
	spListSetParams,
	spListGetParams,
    },
    /* spComponentClassPart */
    {
	SP_TRUE,
	SP_FALSE,
	SP_FALSE,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,

	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
    },
    /* spPrimitiveClassPart */
    {
	0,
    },
    /* spListClassPart */
    {
	0,
    },
};

spComponentClass SpListClass = (spComponentClass)&SpListClassRec;

void spListPartInit(spObject object)
{
    spComponent component = (spComponent)object;

    SpComponentPart(component).spacing_flag = SP_TRUE;

    return;
}

void spListPartFree(spObject object)
{
    return;
}

spBool spListCreate(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    spPrimitiveSetDefaultSize(component);
    if (SpComponentPart(component).disable_initial_scaling == SP_FALSE)
	spPrimitiveScaleDefaultSizeArch(component);
    spListCreateArch(component);
    
    spShowToolTip(component);
    
    if (SpListPart(component).strings != NULL) {
	int i;
	
	for (i = 0;; i++) {
	    if (SpListPart(component).strings[i] == NULL) {
		break;
	    }
	    
	    spDebug(100, "spListCreate", "add strings[%d] = %s\n", i, SpListPart(component).strings[i]);
	    spAddListItem(component, SpListPart(component).strings[i]);
	}

	spDebug(100, "spListCreate", "add strings done: i = %d\n", i);
	    
	if (SpPrimitivePart(component).index >= 0) {
	    spSelectListIndex(component, SpPrimitivePart(component).index);
	}
    }
	
    if (SpComponentPart(component).call_func != NULL) {
	spAddCallback(component, SP_ACTIVATE_CALLBACK | SP_VALUE_CHANGED_CALLBACK,
		      SpComponentPart(component).call_func,
		      SpComponentPart(component).call_data);
    }
    
    return SP_TRUE;
}

spBool spListDestroy(spObject object)
{
    spComponent component = (spComponent)object;
    
    spListDestroyArch(component);
    spPrimitiveDestroyArch(component);
    
    return SP_TRUE;
}

spBool spListSetParams(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    spListSetParamsArch(component);
	
    return SP_TRUE;
}

spBool spListGetParams(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    spListGetParamsArch(component);
	
    return SP_TRUE;
}

int spAddListItem(spComponent component, const char *item)
{
    if (item == NULL) return -1;
    
    if (spIsList(component) == SP_FALSE) {
	if (spIsList(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return -1;
	}
    }

    spDebug(80, "spAddListItem", "item = %s\n", item);
    
    return spAddListItemArch(component, item);
}

int spAddListIndex(spComponent component, const char *item, int index)
{
    if (item == NULL || index < 0) return -1;

    if (spIsList(component) == SP_FALSE) {
	if (spIsList(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return -1;
	}
    }
    
    return spAddListIndexArch(component, item, index);
}

int spDeleteListItem(spComponent component, const char *item)
{
    if (item == NULL) return -1;
    
    if (spIsList(component) == SP_FALSE) {
	if (spIsList(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return -1;
	}
    }

    return spDeleteListItemArch(component, item);
}

int spDeleteListIndex(spComponent component, int index)
{
    if (index < 0) return -1;
    
    if (spIsList(component) == SP_FALSE) {
	if (spIsList(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return -1;
	}
    }
    
    return spDeleteListIndexArch(component, index);
}

int spReplaceListIndex(spComponent component, const char *item, int index)
{
    spBool flag;
    int selected;
    
    if (item == NULL || index < 0) return -1;

    if (spIsList(component) == SP_FALSE) {
	if (spIsList(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return -1;
	}
    }

    selected = spGetSelectedListIndexArch(component);
    
    spDeleteListIndexArch(component, index);
    
    flag = spAddListIndexArch(component, item, index);

    if (selected == index) {
	spSelectListIndex(component, index);
    }

    return flag;
}

spBool spClearListDefault(spComponent component)
{
    while (spDeleteListIndexArch(component, 0) >= 0);

    return SP_TRUE;
}

spBool spClearList(spComponent component)
{
    if (spIsList(component) == SP_FALSE) {
	if (spIsList(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return SP_FALSE;
	}
    }

    return spClearListArch(component);
}

int spFindListItem(spComponent component, const char *item)
{
    if (item == NULL) return -1;
    
    if (spIsList(component) == SP_FALSE) {
	if (spIsList(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return -1;
	}
    }

    return spFindListItemArch(component, item);
}

int spSelectListItem(spComponent component, const char *item)
{
    if (item == NULL) return -1;

    if (spIsList(component) == SP_FALSE) {
	if (spIsList(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return -1;
	}
    }

    return spSelectListItemArch(component, item);
}

int spSelectListIndex(spComponent component, int index)
{
    if (index < 0) return -1;

    if (spIsList(component) == SP_FALSE) {
	if (spIsList(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return -1;
	}
    }

    return spSelectListIndexArch(component, index);
}

char *xspGetSelectedListItem(spComponent component)
{
    if (spIsList(component) == SP_FALSE) {
	if (spIsList(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return NULL;
	}
    }

    return xspGetSelectedListItemArch(component);
}

int spGetSelectedListIndex(spComponent component)
{
    if (spIsList(component) == SP_FALSE) {
	if (spIsList(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return -1;
	}
    }

    SpPrimitivePart(component).index = spGetSelectedListIndexArch(component);
    
    return SpPrimitivePart(component).index;
}

char *xspGetListItem(spComponent component, int index)
{
    if (spIsList(component) == SP_FALSE) {
	if (spIsList(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return NULL;
	}
    }

    return xspGetListItemArch(component, index);
}

int spGetNumListItem(spComponent component)
{
    spDebug(100, "spGetNumListItem", "in\n");
    
    if (spIsList(component) == SP_FALSE) {
	spDebug(100, "spGetNumListItem", "component is not list\n");
    
	if (spIsList(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return -1;
	}
    }

    return spGetNumListItemArch(component);
}

static int spUpDownListIndex(spComponent component, int index, spBool up_flag)
{
    int num_item;
    char *string;
    spBool selected = SP_FALSE;
    
    if ((string = xspGetListItem(component, index)) != NULL) {
	if (index == spGetSelectedListIndex(component)) {
	    selected = SP_TRUE;
	}
	
	if (up_flag == SP_TRUE) {
	    if (index >= 1) {
		spDeleteListIndex(component, index);
		index--;
		spAddListIndex(component, string, index);
	    } else {
		index = -1;
	    }
	} else {
	    num_item = spGetNumListItem(component);
	    if (index < num_item - 1) {
		spDeleteListIndex(component, index);
		index++;
		spAddListIndex(component, string, index);
	    } else {
		index = -1;
	    }
	}
	if (selected == SP_TRUE) {
	    spSelectListIndex(component, index);
	}
	
	xfree(string);
    } else {
	index = -1;
    }

    return index;
}

int spUpListIndex(spComponent component, int index)
{
    return spUpDownListIndex(component, index, SP_TRUE);
}

int spDownListIndex(spComponent component, int index)
{
    return spUpDownListIndex(component, index, SP_FALSE);
}

spBool spIsList(spComponent component)
{
    if (spIsSubClass(component, SpList) == SP_TRUE
	|| spIsSubClass(component, SpComboBox) == SP_TRUE) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spComponent spCreateList(spComponent parent, const char *name, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (spIsCreated(parent) == SP_FALSE || spIsContainer(parent) == SP_FALSE)
	return NULL;
    
    va_start(argp, name);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    return spCreateComponentArg(SpListClass, NULL, name, parent, args, num_arg);
}
