/*
 *	spTabBox_Win.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spMemory.h>

#include <sp/spListP.h>
#include <sp/spDrawP.h>
#include <sp/spImageP.h>
#include <sp/spTabBoxP.h>

#include <uxtheme.h>
#include <vssym32.h>

extern spTopLevel sp_toplevel;

spBool spRedrawTabThemeBackgroundWin(spComponent component)
{
    RECT rect;
    HANDLE htheme;
    HRESULT hr;
    int width, height;
    spBool flag;

    if (spIsOSWinXPLater() == SP_FALSE) {
	return SP_FALSE;
    }

    GetClientRect(SpPrimitiveArch(component).hwnd, &rect);
    spDebug(80, "spRedrawTabThemeBackgroundWin", "client rect: left = %d, top = %d, right = %d, bottom = %d\n",
	    rect.left, rect.top, rect.right, rect.bottom);

    if (!IsRectEmpty(&SpTabBoxArch(component).draw_rect)
	&& EqualRect(&rect, &SpTabBoxArch(component).draw_rect)) {
	spDebug(80, "spRedrawTabThemeBackgroundWin", "rect is not changed: old rect: %d, %d, %d, %d\n",
		SpTabBoxArch(component).draw_rect.left, SpTabBoxArch(component).draw_rect.top,
		SpTabBoxArch(component).draw_rect.right, SpTabBoxArch(component).draw_rect.bottom);
	return SP_FALSE;
    }

    width = rect.right - rect.left;
    height = rect.bottom - rect.top;

    flag = SP_FALSE;
    if (width > 0 && height > 0) {
	CopyRect(&SpTabBoxArch(component).draw_rect, &rect);
	spUpdateCompatibleBitmapWin(component, width, height);

	if ((htheme = OpenThemeData(SpPrimitiveArch(component).hwnd, L"TAB")) != NULL) {
	    spDebug(80, "spRedrawTabThemeBackgroundWin", "OpenThemeData done: htheme = %ld\n", (long)htheme);
	    hr = DrawThemeBackground(htheme, SpPrimitiveArch(component).memdc, TABP_PANE, 0,
				     &rect, NULL);
	    spDebug(80, "spRedrawTabThemeBackgroundWin", "DrawThemeBackground TABP_PANE(%d) result: %ld\n",
		    TABP_PANE, hr);
	    CloseThemeData(htheme);
	    flag = SP_TRUE;
	} else {
	    spDebug(80, "spRedrawTabThemeBackgroundWin", "OpenThemeData failed\n");
	}
    }

    if (flag == SP_FALSE) {
	spFreeDrawingResourcesWin(component);
    }

    return flag;
}

spBool spCopyTabThemeBackgroundWin(spComponent tab_box, spComponent component)
{
    RECT rect;
    int width, height;

    spDebug(80, "spCopyTabThemeBackgroundWin", "in: %s\n", spGetTitle(component));

    if (SpPrimitiveArch(tab_box).hbitmap == NULL || SpPrimitiveArch(tab_box).memdc == NULL) {
	spDebug(10, "spCopyTabThemeBackgroundWin", "tab box doesn't have bitmap\n");
	return SP_FALSE;
    }

    GetWindowRect(SpPrimitiveArch(component).hwnd, &rect);
    OffsetRect(&rect, -rect.left, -rect.top);
    spDebug(80, "spCopyTabThemeBackgroundWin", "original left = %d, top = %d, right = %d, bottom = %d\n",
	    rect.left, rect.top, rect.right, rect.bottom);
    MapWindowPoints(SpPrimitiveArch(component).hwnd, SpPrimitiveArch(tab_box).hwnd, 
		    (POINT *)&rect, 2);
    spDebug(80, "spCopyTabThemeBackgroundWin", "mapped left = %d, top = %d, right = %d, bottom = %d\n",
	    rect.left, rect.top, rect.right, rect.bottom);
    
    width = rect.right - rect.left;
    height = rect.bottom - rect.top;
    spDebug(80, "spCopyTabThemeBackgroundWin", "width = %d, height = %d\n", width, height);
    if (width > 0 && height > 0) {
	spUpdateCompatibleBitmapWin(component, width, height);
	
	spDebug(80, "spCopyTabThemeBackgroundWin", "call BitBlt x = %d, y = %d\n", rect.left, rect.top);
	BitBlt(SpPrimitiveArch(component).memdc, 0, 0, width, height,
	       SpPrimitiveArch(tab_box).memdc, rect.left, rect.top, SRCCOPY);

	return SP_TRUE;
    } else {
	spFreeDrawingResourcesWin(component);
	return SP_FALSE;
    }
}

void spUpdateTabThemeBackgroundBrushWin(spComponent tab_box, spComponent component)
{
    if (SpPrimitiveArch(component).hbrush != NULL) {
	DeleteObject(SpPrimitiveArch(component).hbrush);
	SpPrimitiveArch(component).hbrush = NULL;
    }

    spCopyTabThemeBackgroundWin(tab_box, component);
    spDebug(80, "spUpdateTabThemeBackgroundBrushWin",
	    "hbitmap = %ld\n", (long)SpPrimitiveArch(component).hbitmap);

    if (SpPrimitiveArch(component).hbitmap != NULL) {
	SpPrimitiveArch(component).hbrush = CreatePatternBrush(SpPrimitiveArch(component).hbitmap);

	if (SpPrimitiveArch(component).hbrush == NULL) {
	    spDebug(10, "spUpdateTabThemeBackgroundBrushWin",
		    "**** hbrush == NULL ****, hbitmap = %ld\n", (long)SpPrimitiveArch(component).hbitmap);
	}
    }
    
    spDebug(80, "spUpdateTabThemeBackgroundBrushWin", "done\n");

    return;
}

void spTabBoxPartInitArch(spComponent component)
{
    SetRectEmpty(&SpTabBoxArch(component).draw_rect);
    return;
}

void spTabBoxPartFreeArch(spComponent component)
{
    return;
}

void spTabBoxCreateArch(spComponent component)
{
    SpComponentPart(component).border_width = spScaleX(4);

    /* create tab control */
    SpPrimitiveArch(component).hwnd =
	CreateWindowEx(/*WS_EX_CONTROLPARENT*/0, WC_TABCONTROL,
		       (!strnone(SpGetName(component)) ? SpGetName(component) : ""),
		       WS_CHILD | WS_CLIPSIBLINGS | WS_CLIPCHILDREN
		       | WS_TABSTOP | WS_GROUP | TCS_TABS | spGetVisibleStyleWin(component),
		       SpComponentPart(component).x, SpComponentPart(component).y,
		       SpComponentPart(component).current_width,
		       SpComponentPart(component).current_height,
		       SpParentPrimitiveArch(component).hwnd,
		       (HMENU)SpComponentPart(component).component_id,
		       SpTopLevelArch(sp_toplevel).hThisInst,
		       NULL);
    SendMessage(SpPrimitiveArch(component).hwnd, WM_SETFONT,
		(WPARAM)SpTopLevelArch(sp_toplevel).sys_font, MAKELPARAM(TRUE, 0));

    return;
}

void spTabBoxSetParamsArch(spComponent component)
{
    return;
}

void spTabBoxDestroyArch(spComponent component)
{
    return;
}

int spMapTabItemArch(spComponent component, int index)
{
    return spMapTabItemDefault(component, index);
}

int spGetSelectedTabIndexArch(spComponent component)
{
    return TabCtrl_GetCurSel(SpPrimitiveArch(component).hwnd);
}

void spAddTabItemArch(spComponent component, int index)
{
    TC_ITEM titem;
    spComponent parent;

    parent = SpGetParent(component);
    SpComponentPart(component).top_offset += spScaleY(SP_DEFAULT_TAB_HEIGHT);
    
    titem.mask = TCIF_TEXT | TCIF_IMAGE;
    titem.iImage = -1; 
    titem.pszText = (char *)spGetTitle(component); 

    TabCtrl_InsertItem(SpPrimitiveArch(parent).hwnd, index, &titem);

    return;
}
