/*
 *	spTopLevel_Win.c
 */

#include <windows.h>
#ifdef USE_HTMLHELP
#include <htmlhelp.h>
#endif
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <locale.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spString.h>
#include <sp/spMemory.h>

#include <sp/spLocale.h>

#include <sp/spGraphicsP.h>
#include <sp/spComponentP.h>
#include <sp/spPrimitiveP.h>
#include <sp/spTopLevelP.h>

extern spTopLevel sp_toplevel;

static HINSTANCE sp_hThisInst = NULL;
static HINSTANCE sp_hPrevInst = NULL;
static int sp_nWinMode;
static char *sp_args_buffer = NULL;
static char sp_program_name[SP_MAX_PATHNAME];

#if 1
static BOOL CALLBACK findExecutedApplicationProc(HWND hwnd, LPARAM lParam)
{
    DWORD process_id;
    COPYDATASTRUCT cds;
    LRESULT result;
    spTopLevel *ptoplevel = (spTopLevel *)lParam;

    if (IsWindowVisible(hwnd)) {
	process_id = (DWORD)GetProp(hwnd, SpTopLevelPart(*ptoplevel).unique_id);
	spDebug(100, "findExecutedApplicationProc", "hwnd = %ld, process_id = %ld\n", (long)hwnd, process_id);
    
	if (process_id != 0 && process_id != GetCurrentProcessId()) {
	    cds.dwData = SP_FIND_EXECUTED_APPLICATION_COPYDATA_ID;
	    if (sp_args_buffer != NULL) {
		spDebug(30, "findExecutedApplicationProc", "sending data = %s\n", sp_args_buffer);
		cds.lpData = sp_args_buffer;
		cds.cbData = (DWORD)(strlen(sp_args_buffer) + 1);
	    } else {
		cds.cbData = 0;
		cds.lpData = NULL;
	    }
	    result = SendMessage(hwnd, WM_COPYDATA, (WPARAM)NULL, (LPARAM)&cds);
	    spDebug(30, "findExecutedApplicationProc", "SendMessage result = %ld\n", (long)result);
	    *ptoplevel = NULL;
	
	    return FALSE;
	}
    }
    
    return TRUE;
}

static void findExecutedApplicationWindow(spTopLevel toplevel)
{
    BOOL result;
    
    if (!strnone(SpTopLevelPart(toplevel).unique_id)) {
	result = EnumWindows(&findExecutedApplicationProc, (LPARAM)&toplevel);
	spDebug(30, "findExecutedApplicationWindow", "EnumWindows result = %d\n", result);
	if (toplevel == NULL) {
	    spQuit(0);
	}
    }

    return;
}
#endif

int spWinMainInitialize(HINSTANCE hThisInst, HINSTANCE hPrevInst, int nWinMode)
{
    sp_hThisInst = hThisInst;
    sp_hPrevInst = hPrevInst;
    sp_nWinMode = nWinMode;
    
    GetModuleFileName(sp_hThisInst, sp_program_name,
		      SP_MAX_PATHNAME - 1);
    
    return TRUE;
}

static char *xgetlongname(char *name)
{
    char path[SP_MAX_PATHNAME];
    WIN32_FIND_DATA finddata;
    HANDLE hfind;
        
    if ((hfind = FindFirstFile(name, &finddata)) == INVALID_HANDLE_VALUE) {
	return NULL;
    }
        
#if !defined(__CYGWIN32__)
    {
	char drive[SP_MAX_PATHNAME];
	char dir[SP_MAX_PATHNAME];
	
	strcpy(path, name);
	_splitpath(path, drive, dir, NULL, NULL);
	sprintf(path, "%s%s%s", drive, dir, finddata.cFileName);
    }
#else
    strcpy(path, name);
    if (spGetDirName(path) != NULL) {
	spAddDirSeparator(path);
	strcat(path, finddata.cFileName);
    }
#endif

    FindClose(hfind); 
        
    return strclone(path);
}

#define SP_NUM_ARGV_BUFFER 64

char **spConvertArgsToArgv(LPSTR lpszArgs, char *program_name, int *argcp)
{
    char *p, *name;
    char *args;
    char *string = NULL;
    int count = 0;
    int num_argv_buffer = 0;
    char **argv = NULL;
    
    num_argv_buffer = SP_NUM_ARGV_BUFFER;
    argv = xalloc(num_argv_buffer, char *);

    if (program_name != NULL) {
	argv[count++] = strclone(program_name);
    }

    args = strclone(lpszArgs);

    name = NULL;
    string = args;
    spDebug(30, "spConvertArgsToArgv", "args = %s\n", args);
    while (*string) {
	while (*string) {
	    if (*string != ' ' && *string != '\t') {
		break;
	    }
	    string++;
	}
	if (*string == '"') {
	    string++;
	    name = string;
	    while (*string && *string != '"') string++;
	} else if (*string) {
	    name = string;
	    while (*string) {
		if (*string == ' ' || *string == '\t') {
		    break;
		}
		string++;
	    }
	}
	if (*string) *string++ = NUL;

	if ((p = xgetlongname(name)) == NULL) {
	    p = strclone(name);
	}
	argv[count++] = p;

	if (count >= num_argv_buffer) {
	    num_argv_buffer += SP_NUM_ARGV_BUFFER;
	    argv = xrealloc(argv, num_argv_buffer, char *);
	}
    }
    argv[count] = NULL;

    if (argcp != NULL) *argcp = count;

    xfree(args);
    
    return argv;
}

char **spGetArgv(LPSTR lpszArgs, int *argcp)
{
    static char **argv = NULL;

    if (sp_args_buffer == NULL) {
	sp_args_buffer = strclone((char *)lpszArgs);
    }

    if (argv == NULL) {
	argv = spConvertArgsToArgv(sp_args_buffer, sp_program_name, argcp);
    }
    
    return argv;
}

void spFreeArgv(int argc, char **argv)
{
    int i;
    
    if (argv != NULL) {
	for (i = 0; i < argc; i++) {
	    if (argv[i] == NULL) break;
	    xfree(argv[i]);
	}
	xfree(argv);
    }
}

void spTopLevelPartInitArch(spTopLevel toplevel)
{
    if (sp_hThisInst == NULL) {
	HINSTANCE hinstance;

	if ((hinstance = spGetHInstanceWin()) == NULL) {
	    hinstance = (HINSTANCE)GetModuleHandle(NULL);
	}
	spWinMainInitialize(hinstance, NULL, 0);
    }
    
    return;
}

void spTopLevelPartFreeArch(spTopLevel toplevel)
{
    return;
}

static int helpTranslateMessage(MSG *msg)
{
#ifdef USE_HTMLHELP
    DWORD dwCookie = 0;
    
    if (spGetHelpFileWin(&dwCookie) != NULL && dwCookie != 0) {
	if (HtmlHelp(NULL, NULL, HH_PRETRANSLATEMESSAGE, (DWORD_PTR)msg)) {
	    spDebug(10, "helpTranslateMessage", "receive message\n");
	    return 1;
	}
    }
#endif

    return 0;
}

char *spGetHelpFileWin(DWORD *pdwCookie)
{
#ifdef USE_HTMLHELP
    static DWORD dwCookie = 0;
    static const char *lang = NULL;
    static char help_file[SP_MAX_PATHNAME] = "";

    if (sp_toplevel == NULL || strnone(SpTopLevelPart(sp_toplevel).help_path)) return NULL;

    if (lang == NULL) {
	char id[SP_MAX_LINE];
	char path[SP_MAX_PATHNAME];
	
	lang = spGetLanguage();
	spDebug(10, "spGetHelpFileWin", "lang = %s\n", lang);

	if (spGetLanguageId(lang, id, sizeof(id)) == SP_TRUE) {
	    spDebug(10, "spGetHelpFileWin", "lang id = %s\n", id);
	}

	if (spIsExactName(SpTopLevelPart(sp_toplevel).help_path) == SP_TRUE) {
	    spStrCopy(path, sizeof(path), SpTopLevelPart(sp_toplevel).help_path);
	} else {
	    sprintf(path, "%s%c%s", spGetDefaultDir(), SP_DIR_SEPARATOR,
		    SpTopLevelPart(sp_toplevel).help_path);
	}

	sprintf(help_file, "%s-%s.chm", path, id);
	if (spIsFile(help_file) == SP_FALSE) {
	    spDebug(10, "spGetHelpFileWin", "cannot find help_file: %s\n", help_file);
	    sprintf(help_file, "%s.chm", path);
	    if (spIsFile(help_file) == SP_FALSE) {
		spDebug(10, "spGetHelpFileWin", "cannot find help_file: %s\n", help_file);
		strcpy(help_file, "");
	    }
	}

	if (!strnone(help_file)) {
	    HtmlHelp(NULL, NULL, HH_INITIALIZE, (DWORD_PTR)&dwCookie);
	}
    }

    if (!strnone(help_file)) {
	if (pdwCookie != NULL) *pdwCookie = dwCookie;
	
	return help_file;
    } else {
	return NULL;
    }
#else
    return NULL;
#endif
}

/* reference: http://msdn.microsoft.com/ja-jp/windows/dd630126.aspx */
static void spCalcDPIWin(int *dpiX, int *dpiY)
{
    HDC hdc;

#if 0
    spDebug(30, "spCalcDPIWin", "SM_CXMIN = %d, SM_CYMIN = %d\n",
	    GetSystemMetrics(SM_CXMIN), GetSystemMetrics(SM_CYMIN));
    spDebug(30, "spCalcDPIWin", "SM_CXMINTRACK = %d, SM_CYMINTRACK = %d\n",
	    GetSystemMetrics(SM_CXMINTRACK), GetSystemMetrics(SM_CYMINTRACK));
#endif
    
    if ((hdc = GetDC(NULL)) != NULL) {
	*dpiX = GetDeviceCaps(hdc, LOGPIXELSX);
	*dpiY = GetDeviceCaps(hdc, LOGPIXELSY);
	spDebug(30, "spCalcDPIWin", "dpiX = %d, dpiY = %d\n", *dpiX, *dpiY);
	ReleaseDC(NULL, hdc);	
    }

    return;
}

int spGetDPIXWin(spTopLevel toplevel)
{
    return SpTopLevelArch(toplevel).dpiX;
}

int spGetDPIYWin(spTopLevel toplevel)
{
    return SpTopLevelArch(toplevel).dpiX;
}

int spScaleXWin(spTopLevel toplevel, int x)
{
    return MulDiv(x, SpTopLevelArch(toplevel).dpiX, 96);
}

int spScaleYWin(spTopLevel toplevel, int x)
{
    return MulDiv(x, SpTopLevelArch(toplevel).dpiY, 96);
}

int spUnscaleXWin(spTopLevel toplevel, int x)
{
    return MulDiv(x, 96, SpTopLevelArch(toplevel).dpiX);
}

int spUnscaleYWin(spTopLevel toplevel, int y)
{
    return MulDiv(y, 96, SpTopLevelArch(toplevel).dpiY);
}

int spPointToPixelWin(spTopLevel toplevel, int pt)
{
    return MulDiv(pt, SpTopLevelArch(toplevel).dpiX, 72);
}

void spScaleRectWin(spTopLevel toplevel, RECT *pRect)
{
    pRect->left = spScaleXWin(toplevel, pRect->left);
    pRect->right = spScaleXWin(toplevel, pRect->right);
    pRect->top = spScaleYWin(toplevel, pRect->top);
    pRect->bottom = spScaleYWin(toplevel, pRect->bottom);

    return;
}

int spGetScaledScreenWidthWin(spTopLevel toplevel)
{
    return spUnscaleXWin(toplevel, GetSystemMetrics(SM_CXSCREEN));
}

int spGetScaledScreenHeightWin(spTopLevel toplevel)
{
    return spUnscaleYWin(toplevel, GetSystemMetrics(SM_CYSCREEN));
}

static spBool spGetTaskBarWindow(spTopLevel toplevel)
{
    SpTopLevelArch(toplevel).task_bar = NULL;
    SpTopLevelArch(toplevel).start_button = NULL;

#if 0
    if ((SpTopLevelArch(toplevel).task_bar = FindWindow("Shell_TrayWnd", NULL)) == NULL) {
	return SP_FALSE;
    }

    SpTopLevelArch(toplevel).start_button = FindWindowEx(SpTopLevelArch(toplevel).task_bar, NULL, "Button", NULL);
    if (SpTopLevelArch(toplevel).start_button == NULL) {
	SpTopLevelArch(toplevel).start_button = FindWindow("Button", NULL);
    }
    
    return SP_TRUE;
#else
    return SP_FALSE;
#endif
}

spBool spMapTaskBarWindow(spTopLevel toplevel)
{
#if 0
    if (SpTopLevelArch(toplevel).task_bar != NULL) {
	ShowWindow(SpTopLevelArch(toplevel).task_bar, SW_SHOWDEFAULT);

	if (SpTopLevelArch(toplevel).task_bar != NULL) {
	    ShowWindow(SpTopLevelArch(toplevel).start_button, SW_SHOWDEFAULT);
	}
	
	return SP_TRUE;
    }
#endif

    return SP_FALSE;
}

spBool spUnmapTaskBarWindow(spTopLevel toplevel)
{
#if 0
    if (SpTopLevelArch(toplevel).task_bar != NULL) {
	ShowWindow(SpTopLevelArch(toplevel).task_bar, SW_HIDE);

	if (SpTopLevelArch(toplevel).task_bar != NULL) {
	    ShowWindow(SpTopLevelArch(toplevel).start_button, SW_HIDE);
	}
	
	return SP_TRUE;
    }
#endif

    return SP_FALSE;
}

void spTopLevelCreateArch(spTopLevel toplevel)
{
    WNDCLASSEX wcl;
    INITCOMMONCONTROLSEX ic;
    
    spDebug(30, "spTopLevelCreateArch", "in\n");

    SpTopLevelArch(toplevel).osver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    GetVersionEx(&SpTopLevelArch(toplevel).osver);

    findExecutedApplicationWindow(toplevel);
    
    InitializeCriticalSection(&SpTopLevelArch(toplevel).critical_section);
    
    SpTopLevelArch(toplevel).hmutex = CreateMutex(NULL, FALSE, NULL);
    SpTopLevelArch(toplevel).main_thread_id = GetCurrentThreadId();

    SpTopLevelArch(toplevel).hThisInst = sp_hThisInst;
    SpTopLevelArch(toplevel).hPrevInst = sp_hPrevInst;
    SpTopLevelArch(toplevel).nWinMode = sp_nWinMode;
    SpTopLevelArch(toplevel).sys_font = GetStockObject(DEFAULT_GUI_FONT);
    SpTopLevelArch(toplevel).null_pen = GetStockObject(NULL_PEN);
    SpTopLevelArch(toplevel).null_brush = GetStockObject(HOLLOW_BRUSH);
    SpTopLevelArch(toplevel).bg_brush = GetSysColorBrush(COLOR_BTNFACE);
    SpTopLevelArch(toplevel).hcursor = NULL;
    SpTopLevelArch(toplevel).getting_message = SP_FALSE;
    spCreateGraphics("spTopLevelGraphics",
		     SppForegroundPixel, SP_SYSTEM_COLOR_BLACK_PIXEL,
		     NULL);
    
    spDebug(30, "spTopLevelCreateArch", "spCreateGraphics done\n");

    wcl.hInstance = sp_hThisInst;
    wcl.lpszClassName = SP_WINDOW_CLASS_NAME;
    wcl.lpfnWndProc = spWindowFunc;
    wcl.style = /*CS_HREDRAW | CS_VREDRAW*/0; /* 0 is to prevent from flickering */
    wcl.cbSize = sizeof(WNDCLASSEX);
    wcl.hIcon = NULL; wcl.hIconSm = NULL;
    if (!strnone(SpTopLevelPart(toplevel).icon_name)) {
	wcl.hIcon = LoadIcon(sp_hThisInst, SpTopLevelPart(toplevel).icon_name);
#ifndef _WINE
	if (wcl.hIcon != NULL) {
	    wcl.hIconSm = LoadIcon(sp_hThisInst, SpTopLevelPart(toplevel).icon_name);
	}
#endif
    }
    if (wcl.hIcon == NULL) {
	wcl.hIcon = LoadIcon(NULL, IDI_APPLICATION);
#ifndef _WINE
	wcl.hIconSm = LoadIcon(NULL, IDI_WINLOGO);
#endif
    }
    wcl.hCursor = LoadCursor(NULL, IDC_ARROW);
    wcl.lpszMenuName = NULL;
    wcl.cbClsExtra = 0;
    wcl.cbWndExtra = 0;
    wcl.hbrBackground = /*(HBRUSH)(COLOR_BTNFACE + 1)*/NULL;
    RegisterClassEx(&wcl);

    spDebug(30, "spTopLevelCreateArch", "register window class done\n");
    
    wcl.hInstance = sp_hThisInst;
    wcl.lpszClassName = SP_CANVAS_CLASS_NAME;
    wcl.lpfnWndProc = spWindowFunc;
    wcl.style = 0;
    wcl.cbSize = sizeof(WNDCLASSEX);
    wcl.hIcon = NULL; wcl.hIconSm = NULL;
    if (!strnone(SpTopLevelPart(toplevel).icon_name)) {
	wcl.hIcon = LoadIcon(sp_hThisInst, SpTopLevelPart(toplevel).icon_name);
#ifndef _WINE
	if (wcl.hIcon != NULL) {
	    wcl.hIconSm = LoadIcon(sp_hThisInst, SpTopLevelPart(toplevel).icon_name);
	}
#endif
    }
    if (wcl.hIcon == NULL) {
	wcl.hIcon = LoadIcon(NULL, IDI_APPLICATION);
#ifndef _WINE
	wcl.hIconSm = LoadIcon(NULL, IDI_WINLOGO);
#endif
    }
    wcl.hCursor = LoadCursor(NULL, IDC_ARROW);
    wcl.lpszMenuName = NULL;
    wcl.cbClsExtra = 0;
    wcl.cbWndExtra = 0;
    wcl.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
    RegisterClassEx(&wcl);

    spDebug(30, "spTopLevelCreateArch", "register canvas class done\n");

    ic.dwSize = sizeof(INITCOMMONCONTROLSEX);
    ic.dwICC = 0;
    InitCommonControlsEx(&ic);
    
    spDebug(30, "spTopLevelCreateArch", "InitCommonControls done\n");

    spCalcDPIWin(&SpTopLevelArch(toplevel).dpiX, &SpTopLevelArch(toplevel).dpiY);

    spGetTaskBarWindow(toplevel);

    return;
}
void spTopLevelSetParamsArch(spTopLevel toplevel)
{
    return;
}
    
void spTopLevelDestroyArch(spTopLevel toplevel)
{
    spGraphics graphics, next_graphics;
    spComponent window;
    
    if (toplevel != NULL) {
	spDebug(30, "spTopLevelDestroyArch", "in\n");
	
	/* destroy graphics */
	graphics = SpTopLevelPart(toplevel).graphics;
	while (graphics != NULL) {
	    next_graphics = SpGraphicsPart(graphics).next_graphics;
	    spFreeGraphics(graphics);
	    graphics = next_graphics;
	}

	/* destroy accelerator */
	window = spGetChild(NULL);
	while (window != NULL) {
	    spShowTaskTray(window, SP_FALSE);
	    if (SpPrimitiveArch(window).haccel != NULL) {
		DestroyAcceleratorTable(SpPrimitiveArch(window).haccel);
		SpPrimitiveArch(window).haccel = NULL;
	    }
	    window = SpGetNextComponent(window);
	}

	UnregisterClass(SP_CANVAS_CLASS_NAME, sp_hThisInst);
	UnregisterClass(SP_WINDOW_CLASS_NAME, sp_hThisInst);

	if (SpTopLevelArch(toplevel).hmutex != NULL) {
	    CloseHandle(SpTopLevelArch(toplevel).hmutex);
	}
	
	DeleteCriticalSection(&SpTopLevelArch(toplevel).critical_section);

#ifdef USE_HTMLHELP
	{
	    DWORD dwCookie = 0;
	    
	    if (spGetHelpFileWin(&dwCookie) != NULL && dwCookie != 0) {
		HtmlHelp(NULL, NULL, HH_UNINITIALIZE, (DWORD)dwCookie);
	    }
	}
#endif
	
	spDebug(30, "spTopLevelDestroyArch", "done\n");
    }

    return;
}

void spQuitArch(int status)
{
    spDebug(30, "spQuitArch", "in\n");
    if (sp_toplevel != NULL) {
	spDestroyObject((spObject)sp_toplevel);
    }
    
    return;
}

void spExitArch(int status)
{
    spExit(status);
    return;
}

static int getMessage(spTopLevel toplevel)
{
    BOOL flag;
    
    SpTopLevelArch(toplevel).getting_message = SP_TRUE;
    
    flag = GetMessage(&SpTopLevelArch(toplevel).msg, NULL, 0, 0);
    spDebug(100, "getMessage", "message = %x\n", SpTopLevelArch(toplevel).msg.message);
    
    if (flag && flag != -1) {
	if (!helpTranslateMessage(&SpTopLevelArch(toplevel).msg)
	    && !spTranslateAllAcceleratorWin(&SpTopLevelArch(toplevel).msg)) {
	    
	    spDebug(100, "getMessage", "InSendMessage = %d, thread_id = %ld\n",
		    InSendMessage(), GetCurrentThreadId());
	    
	    TranslateMessage(&SpTopLevelArch(toplevel).msg);
	    DispatchMessage(&SpTopLevelArch(toplevel).msg);

	    spDebug(100, "getMessage", "dispatch done\n");
	}
	
	/*SpTopLevelArch(toplevel).getting_message = SP_FALSE;*/
    
	spDebug(100, "getMessage", "return 1\n");
	
	return 1;
    }
    
    /*SpTopLevelArch(toplevel).getting_message = SP_FALSE;*/
    spDebug(100, "getMessage", "return 0\n");
    
    return 0;
}

int spWaitNextEventArch(spTopLevel toplevel, spBool lock_flag)
{
    return getMessage(toplevel);
}
    
int spDispatchEventArch(spTopLevel toplevel, spBool lock_flag)
{
    if (PeekMessage(&SpTopLevelArch(toplevel).msg, NULL, 0, 0, PM_NOREMOVE)) {
	return getMessage(toplevel);
    } else {
	return 0;
    }
}
    
int spMainLoopArch(spTopLevel toplevel)
{
    while (getMessage(toplevel)) {}
    
    return (int)SpTopLevelArch(toplevel).msg.wParam;
}

spBool spThreadEnterArch(spTopLevel toplevel)
{
    if (SpTopLevelArch(toplevel).hmutex != NULL) {
	spDebug(100, "spThreadEnterArch", "main_thread_id = %ld, thread_id = %ld\n",
		SpTopLevelArch(toplevel).main_thread_id, GetCurrentThreadId());
	if (SpTopLevelArch(toplevel).main_thread_id != GetCurrentThreadId()) {
	    spDebug(100, "spThreadEnterArch", "waiting...\n");
	    WaitForSingleObject(SpTopLevelArch(toplevel).hmutex, INFINITE);
	    spDebug(100, "spThreadEnterArch", "done\n");
	}
    }
    
    return SP_TRUE;
}

spBool spThreadLeaveArch(spTopLevel toplevel)
{
    if (SpTopLevelArch(toplevel).hmutex != NULL) {
	if (SpTopLevelArch(toplevel).main_thread_id != GetCurrentThreadId()) {
	    ReleaseMutex(SpTopLevelArch(toplevel).hmutex);
	    spDebug(100, "spThreadLeaveArch", "mutex released: main_thread_id = %ld, thread_id = %ld\n",
		    SpTopLevelArch(toplevel).main_thread_id, GetCurrentThreadId());
	}
    }
    
    return SP_TRUE;
}

spBool spSuspendMainLockArch(spTopLevel toplevel)
{
    return SP_FALSE;
}

spBool spResumeMainLockArch(spTopLevel toplevel)
{
    return SP_FALSE;
}

spBool spWaitThreadPostDummyEventArch(spTopLevel toplevel)
{
    PostThreadMessage(SpTopLevelArch(toplevel).main_thread_id, WM_NULL, 0, 0);
    
    return SP_TRUE;
}

long spWaitThreadLoopArch(spTopLevel toplevel, void *handle)
{
    return spWaitThreadLoopDefault(toplevel, handle, SP_FALSE);
}

void spEnterCriticalSectionWin(void)
{
    EnterCriticalSection(&SpTopLevelArch(sp_toplevel).critical_section);
    return;
}

void spLeaveCriticalSectionWin(void)
{
    LeaveCriticalSection(&SpTopLevelArch(sp_toplevel).critical_section);
    return;
}
    
spBool spIsPostEventSupportedArch(spTopLevel toplevel)
{
    return SP_TRUE;
}

spBool spIsOSWinNT(void)
{
    if (SpTopLevelArch(sp_toplevel).osver.dwPlatformId == VER_PLATFORM_WIN32_NT) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spIsOSWinXPLater(void)
{
    if (SpTopLevelArch(sp_toplevel).osver.dwMajorVersion > 5
	|| (SpTopLevelArch(sp_toplevel).osver.dwMajorVersion == 5
	    && SpTopLevelArch(sp_toplevel).osver.dwMinorVersion >= 1)) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spIsOSWinVistaLater(void)
{
    if (SpTopLevelArch(sp_toplevel).osver.dwMajorVersion >= 6) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spIsOSWin7Later(void)
{
    if (SpTopLevelArch(sp_toplevel).osver.dwMajorVersion > 6
	|| (SpTopLevelArch(sp_toplevel).osver.dwMajorVersion == 6
	    && SpTopLevelArch(sp_toplevel).osver.dwMinorVersion >= 1)) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}
